/*
 *  sample.dcom.Hub.java
 *
 *  (C) Copyright 1995 - 1999 Microsoft Corporation.  All rights reserved.
 *
 */
package sample.dcom;

import java.util.Hashtable;
import java.util.Enumeration;
/**
 * Java implemented COM object which registers listeners for receiving
 * information sent to this hub through calls to sendMessage and 
 * sendObjects.  This COM object is able to call the methods in the clients
 * directly, as both objects are in the same memory space.  
 */
public class Hub{
	/**
 	 * This is where the listeners are stored.  All listeners will 
 	 * appear as com.ms.com.CUnknown objects to this COM Object.
	 */
	private Hashtable listeners = new Hashtable();
/**
 * Adds the object to the list of objects to be notified for messages 
 * and objects.
 * 
 * @param Object listener - Client implementing sample.IListener
 * 
 * @return String - unique ID for the listener.  This will be needed
 *					to remove the listener from the list.
 */
	public synchronized String addListener(Object listener){
		String key = getKey();
		listeners.put(key,listener);
		System.out.println("Hub - Listener added with key "+key);
		return key;
	}
/**
 * Removes the listener identified from the list of objects to 
 * receive messages and objects.
 * 
 * @param String key - key that was returned when the listener was
 *						added.
 */
	public synchronized void removeListener(String key){
		System.out.println("Hub - removing Listener with key "+key);
		listeners.remove(key);
	}
/**
 * Sends the message to all registered listeners. Unless this COM object
 * was created by a process as a singleton, it will only be possible
 * for a COM object to send messages to itself.
 * 
 * @param String message - message to send to all listeners.
 */
	public synchronized void sendMessage(String message){
		System.out.println("Hub broadcasting message: "+message);
		for (Enumeration e = listeners.elements();e.hasMoreElements();)
			((IListener)e.nextElement()).processMessage(message);
	}
/**
 * Creates a unique key for that will identify a new listener.  
 * 
 * @return String - newly generated key.
 */
	private String getKey(){
		return (new Double(Math.random())).toString();
	}
/**
 * Sends an Array of objects to all registered listeners. Unless this COM object
 * was created by a process as a singleton, it will only be possible
 * for a COM object to send messages to itself.
 * 
 * @param SafeArray safe - contains objects to be broadcast to all listeners. 
 */
	public synchronized void sendObjects(Object[] data){
		System.out.println("Hub sending Objects");
		for (Enumeration e = listeners.elements();e.hasMoreElements();)
			((IListener)e.nextElement()).processObjects(data);
	}
}