/* 
 * 
 * sample.dcom.SingleThreadedApartment.java
 *
 * (C) Copyright 1995 - 1999 Microsoft Corporation.  All rights reserved.
 *
 */
package sample.dcom;

import com.ms.vm.WeakReference;
import com.ms.com.IUnknown;
import com.ms.com.ComLib;
/**
 * This class is used to create instances of Java implemented COM objects on their
 * own Single Threaded Apartment (STA).  The static constructor method will 
 * create a new STA for the given Java class, and instantiate an instance of the
 * java Object on that STA.  The result will be a proxy to that COM object.
 * 
 * A count of the allocated threads is maintained in a global static int. 
 * Monitoring of when the hosted COM object is released is performed by each 
 * thread, and this counter is updated when the COM object is released.  It is 
 * possible to know when all the outstanding COM objects (and threads) are gone, 
 * and it is safe to shutdown your application by monitoring this counter.
 * 
 * 
 * 
 * @version 1.0
 */
public class SingleThreadedApartment {

    // TODO : add support for marshal by value and cloning marshal by value.
    
    /**
     * This is a global counter which specifies the number of allocated STA threads
     * that still have a COM object on them which has a reference count >=1.  
     * Once all of the COM objects have been released, this counter will be zero.
     */
    private static int staThreadCount=0;
	/**
	 * The Thread containing the COM object.
	 */
	public MessageThread object = new MessageThread();
    /**
     * Class that is the COM Object we will host.
     */
    protected Class COMClass = null;
	/**
	 * Thread ID of the Object that has created this apartment
	 */
	private int COMSERVERTHREADID =0;
    /**
     * This object will be used to synchronize our static constructor method
     * with the actual creation of the COM object on its apartment thread.
     */
    protected Object waiter = new Object();
    /**
     * This is the timeout in milliseconds between checks to see if the COM 
     * object we are hosting has been released.
     */
    protected int GC_TIMER=5000;
/**
 * Returns a new COM object that will be serviced by the thread maintained by this
 * host.  Currently a new Thread is created for each new instance of a COM object.
 * 
 * @param Class COMobj - the class which represents the Java COM object we want to
 *          create on its own thread.
 * 
 * @return IUnknown - a proxy to the created COM object.
 */
    public  synchronized static IUnknown createCOMObject(Class COMobj){
        SingleThreadedApartment sta = new SingleThreadedApartment();
        sta.COMClass=COMobj;
	    sta.object.start();
        // We have only started our new STA, and need to wait until the STA has
        // ACTUALLY created the COM object before we attempt to return it!
        synchronized (sta.waiter){
            try{
                  sta.waiter.wait();   
            }catch(InterruptedException i){}
        }
        // We need to remove the instance of the COM Object that we have stored
        // in a local variable, so we first make a copy.
        IUnknown o =  sta.object.comObject;
        // The local instance is cleared, so that our reference tracking will 
        // not be counting our local copy!
        sta.object.comObject=null;
	    return o;
	}
/**
 * Simple constructor that starts the MessageThread that will host the COM object.
 */
	SingleThreadedApartment(){
		COMSERVERTHREADID = com.ms.win32.Kernel32.GetCurrentThreadId();
	}
/**
 * Accessor method for retrieving the count of the number of outstandingn COM
 * objects/STA threads.
 * 
 * @return int - number of outstanding COM objects.
 */
    public static int getSTACount(){
        return staThreadCount;
    }
/**
 * Accessor for incrementing the STA thread count.
 */
    synchronized void incrementSTACount(){
        staThreadCount++;
    }
/**
 * Accessor for decrementing the STA thread count.
 */
    synchronized void decrementSTACount(){
        staThreadCount--;
    }
/**
 * Stops the Windows Timer from sending our STA timer messages, clears our local
 * instance of the served COM object, and calls the garbage collector to do
 * any required cleanup.
 */
	public void quit(){
		com.ms.win32.User32.PostThreadMessage(COMSERVERTHREADID,
											  99,
											this.hashCode(),
											this.hashCode());
		object=null;
		com.ms.lang.SystemX.gc();
	}
/**
 * The actual thread that will host the COM object messages
 */
	class MessageThread extends Thread{
    	/**
		 * Windows thread ID for the apartment
		 */
		public int threadID=0;
		/**
		 * This is the COM object htat we are hosting on this thread.  It cannot be
		 * created here, or it will be attached to the Thread of the class that did a
		 * new on this object.  It must be created at the start of the run method.
		 */
		public WeakReference weakRef = null;
		/**
		 * The proxy to the COM object, which should be returned to who ever needed an
		 * instance of this object.
		 */
		public IUnknown comObject = null;
/**
 * Sets the system to send a message every second to this thread.  This 
 * message will force the thread to check if the COM object has disappeared.
 * 
 * NOTE: This is where the delta between checking if the COM object has gone away
 *       is setup.
 * 
 * @param int id - the window ID for our STA thread, that will receive the WM_TIMER
 *          messages.
 */
		public void setTimer(int id){
			com.ms.win32.User32.SetTimer(id,99,GC_TIMER,null);
		}
/**
 * Stops the timer.
 */
	public void stopTimer(int id){
		com.ms.win32.User32.KillTimer(id,99);

	}
/**
 * Calls the Garbage Collector to ensure that the COM object is cleaned up if all of
 * its references have been released.  Returns true if the COM object is still 
 * aroung after gc (meaning that it still has references) or returns false indicating
 * that the COM object is gone away and it is time to clean up.
 * 
 * @return boolean - whether our COM object is still around.
 */
	public boolean checkCOMObject(){
		com.ms.lang.SystemX.gc();
		if (weakRef.getReferent()==null)
			return false;
		return true;
	}
/**
 * Handles the processing of any windows messages and creates our instance of the COM
 * object.
 */
		public void run(){
            incrementSTACount();
			com.ms.com.ComLib.declareMessagePumpThread();
			threadID = com.ms.win32.Kernel32.GetCurrentThreadId();
			boolean init = false;
            try{
			    //We want to track when the COM object
			    // goes so that we can exit accordingly.
                weakRef = new WeakReference(COMClass.newInstance());
            }catch(Exception i){
                throw new RuntimeException("Unable to create COM object "+i);
            }
 			comObject = (IUnknown)ComLib.makeProxyRef(weakRef.getReferent());
            //We are done creating the Object, so notify our static constructor
            // that it is safe for it to return the object.
	        synchronized (sample.dcom.SingleThreadedApartment.this.waiter){
                sample.dcom.SingleThreadedApartment.this.waiter.notify();
            }
     		// NOTE: we track the actual COM object, not the proxy.
			com.ms.win32.MSG msg = new com.ms.win32.MSG();
		    while (com.ms.win32.User32.GetMessage(msg, 0, 0, 0)) {
				if (!init){//If this is our first time through the loop, we set the 
							// timer.
					init=true;
					setTimer(msg.hwnd);
				}
				if (msg.message==com.ms.win32.win.WM_TIMER){
					if (!checkCOMObject()){
                        decrementSTACount();
                    	com.ms.win32.User32.PostQuitMessage(0);
						stopTimer(msg.hwnd);
						sample.dcom.SingleThreadedApartment.this.quit();// our COM object is gone, so clean up.
                        return;
					}
				setTimer(msg.hwnd);
				}
				com.ms.win32.User32.TranslateMessage(msg);
                com.ms.win32.User32.DispatchMessage(msg);
			}
		}
	}//end of MessageThread class definition.
}//end of class definition.