/*
 *  sample.dcom.COAUTHINFO.java
 *
 *  (C) Copyright 1995 - 1999 Microsoft Corporation.  All rights reserved.
 *
 */
package sample.dcom;

/** 
 * The values of the COAUTHINFO structure determine the authentication settings 
 * used while making a remote activation request from the client's scm to the 
 * server's scm. This structure is defined by default for NTLMSSP, and is 
 * described only for cases that need it to allow DCOM activations to work 
 * correctly with security providers other than NTLMSSP, or to specify 
 * additional security information used during remote activations for 
 * interoperability with alternate implementations of distributed COM. 
 * Currently, the impersonation level must be set to 
 * RPC_C_IMP_LEVEL_IMPERSONATE, or the result will be a failed activation 
 * when the server is running WindowsNT. 
 * 
 * This class will also act as a custom marshaler, it should be used whenever 
 * the COAUTHINFO struct is embedded within another structure.  When used as a 
 * field in another struct, it must have a dll.structmap directive like this:
 * 
 * **
 * * dll.structmap([customMarshal="sample.dcom.COAUTHINFO", type=CUSTOM]) 
 * *
 *  public COAUTHINFO foo;
 * 
 * @see sample.dcom.COAUTHINFO
 * @see sample.dcom.DCOMLib
 * 
 * @version 1.0
 * 
 * @dll.struct(unicode) 
 */
public final class COAUTHINFO implements Cloneable{
    /**
     * The Java VM will read this field to determine how many bytes the custom
     * marshaled object takes up in the native C.
     * 
     * NOTE: static struct fields are NOT considered part of the native C
     *          structure!
     */
    public static int cbByValSize = 4;

/** 
 * A single DWORD value from the list of RPC_C_AUTHN_xxx constants 
 * indicating the authentication service to use. It may be RPC_C_AUTHN_NONE 
 * if no authentication is required.  Currently the only supported value is
 * RPC_C_AUTHN_WINNT, this value should be left as the default.  If no
 * security is required, simply set the Authentication Level to none.  This
 * will effectively be the same as setting this to RPC_C_AUTHN_NONE.
 */
    public int dwAuthnSvc;
/**
 * A single DWORD value from the list of RPC_C_AUTHZ_xxx constants indicating
 * the authorization service to use. If you are using the NT authentication 
 * service, use RPC_C_AUTHZ_NONE. 
 */
    public int dwAuthzSvc;
/** 
 * Pointer to a WCHAR string that indicates the server principal name to use 
 * with the authentication service. If you are using RPC_C_AUTHN_WINNT, the 
 * principal name must be NULL. 
 */
    public String pwszServerPrincName;
/** 
 * A single DWORD value from the list of RPC_C_AUTHN_LEVEL_xxx constants 
 * indicating the authentication level to use. 
 */
	public int dwAuthnLevel;
/** 
 * A single DWORD value from the list of RPC_C_IMP_LEVEL_xxx constants 
 * indicating the impersonation level to use. Currently, only 
 * RPC_C_IMP_LEVEL_IMPERSONATE and RPC_C_IMP_LEVEL_IDENTIFY are supported. 
 * This should be set to RPC_C_IMP_LEVEL_IMPERSONATE to allow the 
 * 'Launch As User' to correctly start a server process as the caller.
 */
	public int dwImpersonationLevel;
/** 
 * Pointer to an AUTH_IDENTITY structure that establishes the identity of the 
 * client. It is authentication-service specific. It is in the form of Windows 
 * NT's SEC_WINNT_AUTH_IDENTITY.
 * 
 * @see com.ms.dcom.COAUTHIDENTITY 
 * 
 * @dll.structmap([customMarshal="sample.dcom.COAUTHIDENTITY", type=CUSTOM]) 
 */
    public COAUTHIDENTITY pAuthIdentity;
/** 
 * A DWORD defining flags to establish indicating the further capabilities of 
 * this proxy. Currently, no capability flags are defined. 
 */
    public int dwCapabilities;
/**
 * Simple constructer for creating an empty COAUTHINFO
 */
    public COAUTHINFO(){}
/**
 * Helps create a populated instance of the COAUTHINFO struct. pwszServerPrincName
 * is set to null, pAuthIdentityData is set to null, and dwCapabilities is 0.
 * dwAuthnSvc is set to RPC_C_AUTHN_WINNT as this is the only Authentication 
 * service supported for Windows NT 4.0 . dwAuthzSvc is set to 0.
 */
    public COAUTHINFO(int dwAuthnLevel,
                      int dwImpersonationLevel,
                      COAUTHIDENTITY coAuthId){
        this.dwAuthnSvc= com.ms.win32.win.RPC_C_AUTHN_WINNT;
        this.dwAuthzSvc = 0;
        this.dwAuthnLevel = dwAuthnLevel;
        this.dwImpersonationLevel = dwImpersonationLevel;
        this.pAuthIdentity=coAuthId;
  }
/**
 * This constructor allows setting of all struct values.
 */      
    public COAUTHINFO(int dwAuthnSvc,
                    int dwAuthzSvc,
                    String pwszServerPrincName,
                    int dwAuthnLevel,
                    int dwImpersonationLevel,
                    COAUTHIDENTITY pAuthIdentityData,
                    int dwCapabilities){
        this.dwAuthnSvc = dwAuthnSvc;
        this.dwAuthzSvc = dwAuthzSvc;
        this.pwszServerPrincName = pwszServerPrincName;
        this.dwAuthnLevel = dwAuthnLevel;
        this.dwImpersonationLevel = dwImpersonationLevel;
        this.pAuthIdentity = pAuthIdentityData;
        this.dwCapabilities = dwCapabilities;
  }
 /**
  * This overrides the default clone method, creating a new instance of this
  * struct with exactly the same values.
  * 
  * @return Object - the new struct with identical values.
  */
    public Object clone(){
        return new COAUTHINFO(dwAuthnSvc,dwAuthzSvc,pwszServerPrincName,
                              dwAuthnLevel,dwImpersonationLevel,pAuthIdentity,
                              dwCapabilities);   
    }
/**
 * The Java VM will call this method to marshal the COAUTHINFO structure from
 * C back to Java.
 * 
 * @param ppStruct - This is a pointer to a memory location where the C structure is.
 * @param Flags - any flags that have been specified by the Java Object representing
 *          the C structure. [Normally this is unused].
 * 
 * @return COAUTHINFO - this is a Java representation of the C structure. This
 *          value will be placed where the custom marshaled field is.
 */    
    public static COAUTHINFO toJava(int ppStruct, int Flags){  
        //This causes a propogate on all of the fields in the COAUTHINFO struct.
        return (COAUTHINFO)com.ms.dll.DllLib.ptrToStruct(COAUTHINFO.class,
                                                com.ms.dll.DllLib.read4(ppStruct));  
    }
/**
 * The Java VM will call this method to marshal the field specified with this 
 * custom marshaler to native C.  This is accomplished by propogating any Java
 * objects to native C, and setting the ppStruct value to point to a pinned 
 * COAUTHINFO struct.
 * 
 * @param COAUTHINFO struct - the struct to be marshaled to native C.
 * @param int Flags - Any flags specified in the dll.structmap declaration for the
 *          field containing the COAUTHINFO object.
 * @param int ppStruct - this is the memory location of where to write the address
 *          of the marshaled COAUTHINFO struct.
 */    
    public static void toExternal(COAUTHINFO struct, int Flags, int ppStruct ) { 
        com.ms.dll.DllLib.propagateStructFields(struct,false);
        int handle = com.ms.dll.Root.alloc(struct);
        com.ms.dll.DllLib.write4(ppStruct,
              com.ms.dll.DllLib.addrOf(handle));
        com.ms.dll.Root.free(handle);
    }
}