/*
 *  sample.dcom.COMServer.java
 *
 *  Copyright (c) 1995 - 1999, Microsoft Corporation
 *
 */
package sample.dcom;

import com.ms.com._Guid;
import com.ms.com.IUnknown;
import com.ms.vm.WeakReference;
/**
 * This class is a simple implementation of a COM Object server, which implements 
 * the COM interface IClassFactory.  It uses a J/Direct call to CoRegisterClass to 
 * add itself to the Running Class Table as the class factory for the given Java
 * implemented COM object, and specified GUID.  Any instances created for the 
 * specified CLSID will cause the SCM to look first at the Running Class Table for
 * a registered Class factory for the given GUID.  SCM will find the ClassFactory 
 * which we have registered, and will then call our CreateInstance method to 
 * create an instance of the COM object.  
 * 
 * Through parameters to the constructor of this class, these options can be
 * specified:
 * 
 *  -    java Class for the that will be server COM object
 *  -    GUID for the COM object
 *  -    If the same object instance should always be returned (singleton)
 *  -    that the served object be hosted in a new single threaded apartment
 *  -    To shutdown the COMServer if all objects have died
 *  -    If the served object should be marshaled by value.
 * 
 * @version 1.1
 */
public class COMServer extends Thread implements com.ms.com.IClassFactory{
    /**
     * Specifies the interval between checks for served COM objects that have
     * died
     */
    protected int GC_TIMER=5000;
	/** 
     * Determines if this server will be serving a Singleton COM object
     */
	private boolean singleton = false;
    /**
     * Determines if we are in the process of shutting down the server.
     */
    private boolean shutdown = false;
    /**
     * This object is used to synchronize the check for shutting down the server
     * with the creation of new objects.
     */
    private Object sync = new Object();
    /**
     * Window Handle used for timer messages.
     */
    int id =0; 
	/**
	 * The ID returned from the CoRegisterClassObject call
	 */
	private int registerID=0;
	/**
	 * GUID that will be used to register this IClassFactory on the Running Class Table.
	 * This is the Guid for the sample.dcom.Hub2 COM object
	 */
	protected  _Guid guid =  null;  
    /**
     * This is the Java Class that represents the COM object that we will be
     * serving.
     */
    protected Class COMClass = null;
    /**
     * Determines if each new instance of the COM object should reside in their
     * own Single Threaded Apartment.
     */
    protected boolean STA = false;
    /**
     * Specifies whether we should shutdown the COMServer if all served objects
     * have been released (no more outstanding objects).
     */
    protected boolean keepAlive = false;
    /**
     * Determines if the served object will be by value
     */
    protected boolean byValue =false;
    /**
     * This is a list of weak references to all outstanding objects.
     */
    protected java.util.Vector servedObjects = new java.util.Vector();
    /**
     * Ensures that we don't shutdown before any objects are served.
     */
    protected boolean started = false;
/**
 * This constructor will Initialize the COMServer with the COM object to be served,
 * the GUID that will be used, and specifications as to whether a singleton, or
 * STA should be used.
 * 
 * @param Class comObject - This is the Java Class that will be used to create an
 *      instance of the COM object to be served whenever someone does a 
 *      CoCreateInstance specifying our GUID 
 * 
 * @param String guid - The GUID for the COM object that we are serving.
 * 
 * @param boolean singleton - determines whether the same COM object will be 
 *          returned for each call to CreateInstace. 
 * 
 * @param boolean singleThreadedApartment - specifies whether the served COM object
 *      (or objects, if the served object is NOT a singleton) should be created
 *      in their own single threaded apartment.
 * 
 * @param boolean keepAlive - Specifies whether this COMServer should shut itself 
 *      down after if there are no outstanding COM objects.  Note - the server
 *      will not start checking until at least one COM object has been given out.
 * 
 * @param boolean byValue - specifies whether this object should be marshaled by
 *      value.  NOTE : THE OBJECT MUST BE SERIALIZABLE!
 */
    public COMServer(Class comObject, 
                     com.ms.com._Guid guid,
                     boolean singleton, 
                     boolean singleThreadedApartment,
                     boolean keepAlive,
                     boolean byValue){
        if (byValue){
            try{
                Object o = comObject.newInstance();
                if ((!(o instanceof java.io.Serializable))&&
                    (!(o instanceof java.io.Externalizable)))
                    throw new RuntimeException("byValue objects MUST be serializable!");
            }catch(Exception ii){
                throw new RuntimeException("Class cannot be instantiated! "+ii);
            }
        }
        this.singleton = singleton;
        this.guid = guid;
        this.keepAlive = keepAlive;
        this.STA = singleThreadedApartment;
        this.COMClass = comObject;
        this.byValue = byValue;
    }
/**
 * Constructor that assumes the object will be served by reference.
 */
    public COMServer(Class comObject,
                     com.ms.com._Guid guid,
                     boolean singleton,
                     boolean singleThreadedApartment,
                     boolean keepAlive){
        this(comObject,guid,singleton,singleThreadedApartment,keepAlive,false);
    }
/**
 * A client application can call this method to determine which GUID we are 
 * serving COM objects for.
 * 
 * @return String - GUID for the COM object we are serving.
 */
    public String getCLSID(){
        if (guid==null)
            return null;
        return guid.toString();
    }
/**
 * This method obtains the instance of the COM object that will be given out to 
 * service the current request.  It checks the singleton, and STA specifications
 * and acts accordingly.  
 * 
 * @return IUnknown - newly created COM object, on its own apartment thread - or
 *          the singleton COM object, depending on global initialization params.
 */
  	protected IUnknown getCOMObject()throws Exception{
        if (singleton)
            return getSingleton();   
        Object hostObject = null;
        if (STA){
            //Create Single Threaded Apartment Object
            hostObject = SingleThreadedApartment.createCOMObject(COMClass);
        }else{
            //Create Free Threaded object
            if (byValue)
                hostObject = sample.marshal.MarshalLib.marshalByValue(COMClass.newInstance());
            else
                hostObject = COMClass.newInstance();
            System.out.println("Hard hold last served");
        }
        //Store weak reference to hostObject for tracking.
        this.servedObjects.addElement(new WeakReference(hostObject));
        if (STA)
            return (IUnknown)hostObject;
        else 
            return (IUnknown)com.ms.com.ComLib.makeProxyRef(hostObject);
 	}
/**
 * This method will obtain the singleton object that we are serving.  It will 
 * first check to see if we have already created a singleton and whether it is
 * still alive.  If there is no available object, a new one is created.  The 
 * single threaded apartment parameter to the COMServer constructer is used 
 * to determine how the new COM Object instance should be created.
 * 
 * @return com.ms.com.IUnknown - this is the Singleton object we are serving.
 * 
 * @exception Exception - This may arive from creating a COM object, or from 
 *          attempting to create the object on an STA.
 */
    protected IUnknown getSingleton() throws Exception{
        // TODO : if we are marshal by value and singleton, we should return a clone
        
        //Check if an instance of our singleton has already been created.
        Object holder = null;
        if (this.servedObjects.size()!=0){
            // We hold a weak reference to a singleton
            WeakReference weak = (WeakReference)servedObjects.firstElement();
            holder = weak.getReferent();
            if (holder!=null)
                return (IUnknown)holder;
            else{//Instance of Object has died, need a new one.
                if (byValue)
                    holder = sample.marshal.MarshalLib.marshalByValue(COMClass.newInstance());
                else 
                    holder = this.COMClass.newInstance();
                weak.setReferent(holder);
                return (IUnknown)holder;
            }
        }else{// We have no weak references to COM objects, need to create new.
            if (this.STA)
                holder = SingleThreadedApartment.createCOMObject(COMClass);
            else
                if (byValue)
                    holder = sample.marshal.MarshalLib.marshalByValue(COMClass.newInstance());
                else    
                    holder = COMClass.newInstance();
            WeakReference weak = new WeakReference(holder);
            servedObjects.addElement(weak);
            com.ms.lang.SystemX.gc();
            if (this.STA)
                return (IUnknown)holder;
            else
                return (IUnknown)com.ms.com.ComLib.makeProxyRef(holder);
        }
    }
/**
 * Determines if there are anymore apartment threads running, if not shuts us down.
 */
	private void checkShutdown(){
        synchronized (sync){
            com.ms.lang.SystemX.gc();
            boolean dead=false;
            if (this.STA){
                if(SingleThreadedApartment.getSTACount()==0)
                    dead=true;
            }else{
                int alive =0;
                for (int i=0;i<this.servedObjects.size();i++)
                    if (((WeakReference)servedObjects.elementAt(i)).getReferent()!=null)
                        alive++;
              if ((alive==0)&&(!this.keepAlive))
                    dead=true;
            }//end else
            if (dead){
                shutdown();
                try{Thread.sleep(1000);}catch(Exception ii){}
                System.exit(0);
   	        }
        }//end sync
	}
/**
 * This method performs all of the cleanup for this COMServer.  It removes the 
 * IClassFactory from the Running Object table, and calls 
 * com.ms.com.ComLib.release() on all outstanding served objects.
 */
    public synchronized void shutdown(){
        synchronized (sync){
			try{
            shutdown=true;
	// 
	// We may be called from a thread that has NOT been initialized for COM 
	// use as an MTA. To cover for this case, need to call CoRevokeClassObject
	// from an MTA thread. Easiest way to do this is to spin one up for the job.
	//
			Thread t = new Thread(){
				public void run(){
			// Remove this IClassFactory from the Running ClassObject Table
						com.ms.win32.Ole32.CoRevokeClassObject(registerID);
				}};
			com.ms.com.ComLib.startMTAThread(t);
			t.join();
			for (int i=0;i<this.servedObjects.size();i++){
                Object o = 
                       ((WeakReference)servedObjects.elementAt(i)).getReferent();
                if (o!=null)
                    com.ms.com.ComLib.release(o);
            }   
			}catch(Exception ii){
				
				System.out.println(ii);
			}
        }
    }
/**
 * Ensures that the Class Factory cleans up after itself it is about to die..
 */
    public void finalize(){
		if (shutdown==false){//we Haven't already shutdown   
			shutdown();
		}
    }
/**
 * This method is implemented as part of the IClassFactory interface, it is usually
 * used to optimize the creation of COM objects.  It currently does nothing in this
 * implementation.  E_NOTIMPL is returned if it is called.
 * 
 * @param int flock
 */
	public synchronized void LockServer(boolean flock){
        debug("Inside LockServer "+flock);
        throw new com.ms.com.ComFailException(com.ms.win32.win.E_NOTIMPL);
	}
/**
 * This method is implemented as part of the IClassFactory interface, it is called to
 * create an instance of the object, and should set the IUnknown to a pointer to the
 * interface requested (refid) on the served object.
 * 
 * @param int unused - does nothing
 * @param _Guid refid - CLSID of the interface requested.
 * @param IUnknown[] punk - pointer which needs to be set to the requested interface. 
 */
	public synchronized com.ms.com.IUnknown CreateInstance(
                                             com.ms.com.IUnknown pUnkOuter, 
                                            _Guid refid){
        debug("Inside CreateInstance");
        if (pUnkOuter!=null)
            throw new com.ms.com.ComFailException(
               com.ms.win32.win.CLASS_E_NOAGGREGATION);
        synchronized (sync){
            // We may have been waiting for the COMServer to shutdown, so check
            // if we are still open for business.
            if (shutdown)
                throw new com.ms.com.ComFailException(0x80040154);
            try{
                debug("Returning COM Object");
                started=true;
                return getCOMObject();
            }catch(Exception ii){
                debug("Exception calling getCOMObject() throwing E_UNEXPECTED "+ii);
                throw new com.ms.com.ComFailException(com.ms.win32.win.E_UNEXPECTED);
            }
        }//end sync
	}
/**
 * Registers this Object on the Running Class Table, using the CoRegisterClassObject
 * method in com.ms.win32.Ole32.  This method then goes into a message loop, which
 * is periodically stimulated by a win32 timer to allow the server to check for
 * served COM objects which have been released.
 */

	public void run(){
        registerID = com.ms.win32.Ole32.CoRegisterClassObject(guid,
                                               this,
                                               com.ms.win32.win.CLSCTX_SERVER, 
                                               com.ms.win32.win.REGCLS_MULTIPLEUSE);
        while (!shutdown){
            try{Thread.sleep(GC_TIMER);}catch(Exception ii){}
            checkShutdown();
        }
	}
/**
 * Simple method for controlling the debugging output from this class.
 * 
 * @param String Description text from debug line
 */
    protected void debug(String data){
        System.out.println(data);   
    }
}