/*
 * sample.ntservice.NTService.java
 *
 * (C) Copyright 1995 - 1999 Microsoft Corporation.  All rights reserved.
 *
 */
package sample.ntservice;
/**
 * This generic implementation of an NT Service using the J/Direct methods from
 * win32 API Advapi32.dll contains methods for registering/removing the service, 
 * writting to the NT Event log, and for sending status to the SCM.  Creation of a 
 * custom NT Service is accomplished by subclassing this abstract class, and 
 * overriding the abstract methods.  Additional configurable parameters (which have
 * default values) can be overridden to specify load order, dependencies and 
 * installation/removal parameters.
 * 
 * Supports these command line arguments in the init method:
 * 
 *  -   install         Installs the NT Service on the local machine, adding
 *                      all necessary registry keys.  The current location of the
 *                      .exe file is assumed to be in the correct place.
 * 
 *  -   remove          Remove the NT Service from the system, and all associated
 *                      registry keys.
 * 
 *  -   debug           Runs the Service in a command window, Debug mode.
 * 
 * @see sample.ntservice.Advapi32
 * @see sample.util.StructUtil
 * 
 * @version 1.0
 */
public abstract class NTService{        
        /**
         * Object used to synchronize the stop event from the ServiceControl object
         * which may be called from a seperate thread from the ServiceMain.
         */
        Object stopevt = new Object();
        /**
         * This structure will be used repeatedly by us to communicate with the 
         * SCM.
         */
		SERVICE_STATUS ssStatus = new SERVICE_STATUS();
        /**
         * This handle is used repeatedly for sending status to the SCM.
         */
		SERVICE_STATUS_HANDLE sshStatusHandle = null;
        /**
         * This is a flag used in Advapi32 function calls that is not currently
         * available in com.ms.win32.win , it is essentially an OR of all the 
         * SERVICE_* access flags.
         */
        public static final int SERVICE_ALL_ACCESS = 
                            com.ms.win32.win.SERVICE_CHANGE_CONFIG|
                            com.ms.win32.win.SERVICE_ENUMERATE_DEPENDENTS|
                            com.ms.win32.win.SERVICE_INTERROGATE|
                            com.ms.win32.win.SERVICE_PAUSE_CONTINUE|
                            com.ms.win32.win.SERVICE_QUERY_CONFIG|
                            com.ms.win32.win.SERVICE_QUERY_STATUS|
                            com.ms.win32.win.SERVICE_START|
                            com.ms.win32.win.SERVICE_STOP|
                            com.ms.win32.win.SERVICE_USER_DEFINED_CONTROL|
                            com.ms.win32.win.STANDARD_RIGHTS_REQUIRED;
	    /**
         * This is a flag used in Advapi32 function calls that is not currently
         * available in com.ms.win32.win , it is essentially an OR of all the 
         * SC_* access flags.
	     */
        public static final int SC_MANAGER_ALL_ACCESS =
                            com.ms.win32.win.SC_MANAGER_CONNECT|
                            com.ms.win32.win.SC_MANAGER_CREATE_SERVICE|
                            com.ms.win32.win.SC_MANAGER_ENUMERATE_SERVICE|
                            com.ms.win32.win.SC_MANAGER_LOCK|
                            com.ms.win32.win.SC_MANAGER_QUERY_LOCK_STATUS|
                            com.ms.win32.win.STANDARD_RIGHTS_REQUIRED;
/**
 * This method should contain all of the code that needs to be executed if the 
 * system or a user needs to pause the implemented service.
 * 
 * @see sample.ntservice.NTService.ServiceControl
 */
    protected abstract void servicePause();
/**
 * This method must contain all of the code that needs to be executed to process
 * the current service from a paused to a running state.
 * 
 * @see sample.ntservice.NTService.ServiceControl
 */
    protected abstract void serviceContinue();
/**
 * All code contained in this method will be executed when the service is stopped
 * by the system or by a user.  This method should also perform any cleanup of 
 * variables, connection or state - and release any held resources.  The length of
 * time the system will wait for the execution of this method needs to be specified
 * in getServiceStopTimeout.  If the method takes longer than this time to complete, 
 * the system will assume that there was a problem stopping the service, and throw
 * and exception.
 * 
 * @see #getServiceStopTimeout
 * @see sample.ntservice.NTService.ServiceControl
 */
    protected abstract void serviceStop();
/**
 * The startup code for the user implemented service should be placed in this method.
 * Any command line arguments that are specified during startup will be passed as
 * parameters to this method, NOTE that these are coming from the C world, so they
 * appear an integer specifying the number of arguments and a pointer to the actual
 * arguments.  The amount of time that this method reauires must be specified by the
 * getServiceStartTimeout method, which is used to determine how long the SCM must
 * wait for the startup to finish.  If the start method takes longer than this amount
 * of time, the system will assume that the start was a failure, and throw an
 * exception.
 * 
 * @see #getServiceStartTimeout
 * @see sample.ntservice.NTService.ServiceMain
 */
    protected abstract void serviceStart(int dwArgc, int lpszArgv);
/**
 * This method will be invoked when the service is informed that the system is
 * shutting down.  The service will be stopped AFTER this method is invoked.
 * 
 * @see sample.ntservice.NTService.ServiceControl
 */
    protected void serviceShutdown(){}
/**
 * This will returnthe amount of time required by the serviceStart method to complete.
 * It will be used to determine when the startup has failed, by telling the SCM to 
 * wait this long before the status of the startup is reported.
 * 
 * @see #serviceStart
 * @see sample.ntservice.NTService.ServiceMain
 */
    protected abstract int getServiceStartTimeout();
/**
 * This will returnthe amount of time required by the serviceStop method to complete.
 * It will be used to determine when the stop has failed, by telling the SCM to 
 * wait this long before the status of the stop is reported.
 * 
 * @see #serviceStop
 * @see sample.ntservice.NTService.ServiceControl
 * @see sample.ntservice.NTService.ServiceMain
 */
    protected abstract int getServiceStopTimeout();
/**
 * This method must return a string that names the service to install. The maximum 
 * string length is 256 characters. The service control manager database preserves 
 * the case of the characters, but service name comparisons are always case 
 * insensitive. Forward-slash (/) and back-slash (\) are invalid service name 
 * characters.
 * 
 * @return String - String that names the service.
 * 
 * @see #register
 */
    protected abstract String getServiceName();
/**
 * This method must return a string that is to be used by user interface programs 
 * to identify the service. This string has a maximum length of 256 characters. The 
 * name is case-preserved in the service control manager. display name comparisons 
 * are always case-insensitive. 
 * 
 * @return String - String that represents the display name of this service.
 * 
 * @see #register
 */
    protected abstract String getServiceDisplayName();
/**
 * Retrieves a double null-terminated array of null-separated names of services or 
 * load ordering groups that the system must start before this service. Specify NULL
 * or an empty string if the service has no dependencies. Dependency on a group 
 * means that this service can run if at least one member of the group is running 
 * after an attempt to start all members of the group. 
 * 
 * You must prefix group names with SC_GROUP_IDENTIFIER so that they can be 
 * distinguished from a service name, because services and service groups share the 
 * same name space. 
 * 
 * NOTE the default value for this method is null.
 * 
 * @return Object - Object that represents the dependencies for this service.
 * 
 * @see #register
 * @see sample.ntservice.Advapi32.CreateService
 */
    protected Object getServiceDependencies(){
        return null;
    }
/**
 * Retrieves a string that names the load ordering group of which this service is a 
 * member. Specify NULL or an empty string if the service does not belong to a group. 
 * 
 * NOTE the default value for this method is null
 * 
 * @return String - Represents the load order group for this service.
 * 
 * @see #register
 * @see sample.ntservice.Advapi32.CreateService
 */
    protected String getServiceLoadOrderGroup(){
        return null;
    }
/**
 * Retrieves a set of bit flags that specify the type of service. 
 * 
 * @return int - Must specify one of the following service types. 
 * 
 *          Value                           Meaning
 * 
 *          SERVICE_WIN32_OWN_PROCESS       Specifies a Win32-based service that 
 *                                          runs in its own process. 
 * 
 *          SERVICE_WIN32_SHARE_PROCESS     Specifies a Win32-based service that 
 *                                          shares a process with other services. 
 * 
 *          SERVICE_KERNEL_DRIVER           Specifies a driver service. 
 * 
 *          SERVICE_FILE_SYSTEM_DRIVER      Specifies a file system driver service. 
 * 
 *          If you specify either SERVICE_WIN32_OWN_PROCESS or 
 *          SERVICE_WIN32_SHARE_PROCESS, you can also specify the following flag. 
 *          
 *          Value                           Meaning 
 * 
 *          SERVICE_INTERACTIVE_PROCESS     Enables a Win32-based service process to
 *                                          interact with the desktop. 

 */
    protected int getProcessType(){
        return com.ms.win32.win.SERVICE_WIN32_OWN_PROCESS;
    }
/**
 * Retrieves a string that names the target computer when removing a service. If the
 * return value is NULL or points to an empty string, the function connects to the 
 * service control manager on the local computer. 
 * 
 * NOTE: the default return value is null.
 * 
 * @return String - representing the machine name.
 * 
 * @see #remove
 * @see sample.ntservice.Advapi32.OpenSCManager
 * @see sample.ntservice.Advapi32.DeleteService
 */
    protected String getMachineName(){
        return null;
    }
/**
 * Retrieves a string that names the service control manager database to open. This 
 * should be set to the SERVICES_ACTIVE_DATABASE. If the return value is NULL, the 
 * SERVICES_ACTIVE_DATABASE database is opened by default.
 * 
 * NOTE: the default return value is null.
 *  
 * @return String - represents the SCM Database.
 * 
 * @see #remove
 * @see sample.ntservice.Advapi32.OpenSCManager
 */
    protected String getDatabaseName(){
        return null;
    }
/**
 * Retrieves the specification for when to start the service. 
 * 
 * @return int - The return value must be one of the following start types. 
 * 
 *          Value                           Meaning
 * 
 *          SERVICE_BOOT_START              Specifies a device driver started by the
 *                                          system loader. This value is valid only 
 *                                          for driver services. 
 * 
 *          SERVICE_SYSTEM_START            Specifies a device driver started by the 
 *                                          IoInitSystem function. This value is 
 *                                          valid only for driver services. 
 * 
 *          SERVICE_AUTO_START              Specifies a service to be started 
 *                                          automatically by the service control 
 *                                          manager during system startup. 
 * 
 *          SERVICE_DEMAND_START            Specifies a service to be started by the 
 *                                          service control manager when a process 
 *                                          calls the StartService function. 
 * 
 *          SERVICE_DISABLED                Specifies a service that can no longer 
 *                                          be started. 
 */
    protected int getStartupType(){
        return com.ms.win32.win.SERVICE_DEMAND_START;
    }
/**
 * Retrieves the specification for the severity of the error if this service fails 
 * to start during startup, and determines the action taken by the startup program 
 * if failure occurs. 
 * 
 * @return int - Must specify one of the following error control flags.
 * 
 *          Value                           Meaning
 * 
 *          SERVICE_ERROR_IGNORE            The startup program logs the error but 
 *                                          continues the startup operation. 
 * 
 *          SERVICE_ERROR_NORMAL            The startup program logs the error and 
 *                                          puts up a message box pop-up but 
 *                                          continues the startup operation. 
 * 
 *          SERVICE_ERROR_SEVERE            The startup program logs the error. If 
 *                                          the last-known-good configuration is 
 *                                          being started, the startup operation 
 *                                          continues. Otherwise, the system is 
 *                                          restarted with the last-known-good 
 *                                          configuration. 
 * 
 *          SERVICE_ERROR_CRITICAL          The startup program logs the error, if 
 *                                          possible. If the last-known-good 
 *                                          configuration is being started, the 
 *                                          startup operation fails. Otherwise, the 
 *                                          system is restarted with the last-known 
 *                                          good configuration. 
 */
    protected int getErrorType(){
        return com.ms.win32.win.SERVICE_ERROR_NORMAL;
    }
/**
 * Retrieves a tag value that is unique in the group specified in the 
 * LoadOrderGroup. This will return NULL if you are not changing the existing tag. 
 * 
 * This tag can be used for ordering service startup within a load ordering group by
 * specifying a tag order vector in the GroupOrderList value of the following 
 * registry key: 
 * 
 *          HKEY_LOCAL_MACHINE\System\CurrentControlSet\Control 
 * 
 * Tags are only evaluated for driver services that have SERVICE_BOOT_START or 
 * SERVICE_SYSTEM_START start types. 
 * 
 * NOTE: null is the default return value for this method.
 * 
 * @return int - tagID
 * 
 * @see #getServiceLoadOrderGroup
 * @see sample.ntservice.Advapi32#CreateService
 */
    protected String getServiceLoadOrderTagID(){
        return null;
    }
/**
 * If the service type is SERVICE_WIN32_OWN_PROCESS, use an account name in the form
 *  DomainName\UserName will be returned. The service process will be logged on as 
 * this user. If the account belongs to the built-in domain, you can specify 
 * .\UserName. If the service type is SERVICE_WIN32_SHARE_PROCESS the return value 
 * must be  the LocalSystem account. If NULL is returned the LocalSystem account. 
 * 
 * If the service type is SERVICE_KERNEL_DRIVER or SERVICE_FILE_SYSTEM_DRIVER, the 
 * name is the driver object name that the system uses to load the device driver. 
 * Specify NULL if the driver is to use a default object name created by the I/O 
 * system.
 * 
 * NOTE: the default return value is null.
 * 
 * @return String - represents the acount to be used for running this service 
 * 
 * @see #getServicePassword
 * @see sample.ntservice.Advapi32#CreateService
 */
    protected String getServiceDomainUser(){
        return null;
    }
 /**
 * Retrieves the password to the account name specified by getServiceDomainUser().
 * If the return value is NULL or an empty string, the service has no password. 
 * Passwords are ignored for driver services. If getServiceDomainUser() is 
 * LocalSystem, the password must be NULL.
 * 
 * NOTE: The default return value is null.
 * 
 * @return String - represents the password used for the account hositng the service.
 * 
 * @see #getServiceDomainUser
 * @see sample.ntservice.Advapi32.CreateService
 */
    protected String getServicePassword(){
        return null;
    }
/**
 * Default initialization method that takes the command line parameters and 
 * evaluates them to determine if the service should be installed, removed or
 * started.  Subclasses of NTService should implement their own main methods and
 * call this method with the command line parameters.
 *
 * @param String[] args - command line arguments.  The only arguments used are:
 * 
 *          "-install" to install the service on the local machine
 * 
 *          "-remove" removes the service from the local machine.
 * 
 *          If no parameters are given, or they are not one of the above, this 
 *          method will attempt to start the service.  NOTE: this will fail unless
 *          the SCM is trying to start this service! (I.E the service cannot be
 *          started from the command line using the service .exe file) 
 */
    protected void init(String[] args)throws Exception{
             SERVICE_TABLE_ENTRY[] serviceDispatchTable = 
                    (SERVICE_TABLE_ENTRY[])sample.util.StructArray.newStructArray(
                                            SERVICE_TABLE_ENTRY.class,2);
  		    serviceDispatchTable[0].lpServiceName =getServiceName();
		    serviceDispatchTable[0].lpServiceProc =com.ms.dll.DllLib.addrOf(
                           com.ms.dll.Root.alloc(new ServiceMain()));
		    serviceDispatchTable[1].lpServiceName=null;
		    serviceDispatchTable[1].lpServiceProc = 0;
            if (!Advapi32.StartServiceCtrlDispatcher(
                                      sample.util.StructArray.toCStructArray(
                                                        serviceDispatchTable))){
            this.logEvent(null,new String[]{"Start Service Ctrl Dispatch Error "+
							com.ms.dll.DllLib.getLastWin32Error()},
                          (short)com.ms.win32.win.EVENTLOG_ERROR_TYPE,
                          null,2000,3);
		    }
	}
/**
 * Removes this service from the local system, by first stopping the service, and
 * removing all related entries from the SCM database.  Information regarding the
 * progress of the removal is written to standard out.
 * 
 * getMachineName() and getDatabaseName() will be called to get the machine and 
 * database for which this service will be removed from.
 * 
 * @see sample.ntservice.Advapi32.DeleteService
 * @see #getMachineName
 * @see #getDatabaseName
 */
    protected void remove(){
        SERVICE_STATUS[] ssStatus = new SERVICE_STATUS[1];
        ssStatus[0] = new SERVICE_STATUS();
        SC_HANDLE schSCManager = Advapi32.OpenSCManager(
                                            getMachineName(),
                                            getDatabaseName(),
                                            SC_MANAGER_ALL_ACCESS);
        SC_HANDLE schService = Advapi32.OpenService(
                                            schSCManager,
                                            getServiceName(),
                                            SERVICE_ALL_ACCESS);
        if (schService!=null){
        // try to stop the service
            Advapi32.ControlService(schService,
                                com.ms.win32.win.SERVICE_CONTROL_STOP,
                                ssStatus[0]);
            try{Thread.sleep(1000);}catch(Exception ii){}
            while (Advapi32.QueryServiceStatus(schService,ssStatus[0])){
                if (ssStatus[0].dwCurrentState==com.ms.win32.win.SERVICE_STOP_PENDING){
                    System.out.print(".");
                    try{Thread.sleep(1000);}catch(Exception ii){}
                }else
                    break;
            }
            if (ssStatus[0].dwCurrentState==com.ms.win32.win.SERVICE_STOPPED)
                System.out.println("\n"+getServiceDisplayName()+" stopped.");
            else
                System.out.println(this.getServiceDisplayName()+" failed to stop.");
            // now remove the service
            if (Advapi32.DeleteService(schService))
                System.out.println(getServiceDisplayName()+" removed.");
            else 
                System.out.println("DeleteService failed - "+com.ms.dll.DllLib.getLastWin32Error());
            Advapi32.CloseServiceHandle(schService);
       }else
            System.out.println("\nUnable to remove the service, Service not registered!");
       Advapi32.CloseServiceHandle(schSCManager);
    }
/**
 * This method creates a service object and installs it in the service control 
 * manager database by creating a key with the same name as the service under the 
 * following registry key:
 * 
 *      HKEY_LOCAL_MACHINE\System\CurrentControlSet\Services 
 * 
 * Information specified by this method is saved as values under this key. Setup 
 * programs and the service itself can create subkeys under this key for any service
 * specific information. 
 * 
 * The getMachineName() and getDatabaseName() methods will be called to get the 
 * machine and database that will be the target of this installation.
 * 
 * @see #getMachineName
 * @see #getMachineName
 * @see sample.ntservice.Advapi32.OpenSCManager
 * @see sample.ntservice.Advapi32.CreateService
 */
	protected void register(){
        if (this.getModuleFileName().indexOf("jview")!=-1){
           System.out.println("\nThe Service must be an .exe file!");
           System.exit(1);
        }
  		SC_HANDLE sc= Advapi32.OpenSCManager(this.getMachineName(),
										   this.getDatabaseName(),
										   com.ms.win32.win.GENERIC_WRITE);
		SC_HANDLE foo = Advapi32.CreateService(sc,
							   getServiceName(),
							   getServiceDisplayName(),
          					   SERVICE_ALL_ACCESS,
							   getProcessType(),
                               getStartupType(),
                               getErrorType(),
							   this.getModuleFileName(),
							   this.getServiceLoadOrderGroup(),
							   this.getServiceLoadOrderTagID(),
							   this.getServiceDependencies(),
                               this.getServiceDomainUser(),
							   this.getServicePassword());
		if (foo==null)
			System.out.println("\nRegistration failed!");
        else 
            System.out.println("\nRegistration succeeded!");
		Advapi32.CloseServiceHandle(sc);
	}
/**
 * This method uses the Kernel32 function 'GetModuleFileName' to obtain the
 * full path name of the executable that this method is running in.  This is
 * useful for determining registry key values which require a pathname to the
 * service executable.
 * 
 * @return String - full path name of the executable containing this service.
 */
    protected String getModuleFileName(){
        StringBuffer filename = new StringBuffer(512);
        com.ms.win32.Kernel32.GetModuleFileName(0,filename,512);
        return filename.toString();
    }
/**
 * This method is used to add registry keys for the message file, so that our
 * service will be able to write our defined messages to the event log, and that
 * the event log will be able to decipher those messages.
 *
 *  HKEY_LOCAL_MACHINE\System\CurrentControlSet\Services\EventLog\Application
 *                                          \<Our Service Name>
 * 
 *  
 * @param String eventPath - full pathname to the exe or dll which contains the 
 *          message resource required for formating event log messages.
 * 
 * @param eventTypes - value used for the TypesSupported value.
 * 
 * @param String categoryPath - full pathname to the exe or dll which contains the
 *          category resource information.  Typically this would be the same as 
 *          the event message file.
 * 
 * @param int catCount - the number of catagories defined for this event log.  
 *          These categories MUST be number from 0 to this number -1 in the
 *          message file.
 * 
 * @exception com.ms.lang.RegKeyException - thrown if there was a problem dealing
 *          with these keys in the registry.
 */
    protected void insertRegistryKeys(String eventPath,
                                      int eventTypes,
                                      String categoryPath,
                                      int catCount) throws com.ms.lang.RegKeyException{
            com.ms.lang.RegKey myKey = new com.ms.lang.RegKey(
                            com.ms.lang.RegKey.LOCALMACHINE_ROOT,
  "SYSTEM\\CurrentControlSet\\Services\\EventLog\\Application\\"+
    this.getServiceName(),
                            com.ms.lang.RegKey.KEYOPEN_CREATE);
            
            myKey.setValue("EventMessageFile",eventPath);
            myKey.setValue("TypesSupported",eventTypes);
            myKey.setValue("CategoryMessageFile",categoryPath);
            myKey.setValue("CategoryCount",catCount);
    }
/**
 * Removes all registry keys that deal with the eventlog.  These are under:
 * 
 *  HKEY_LOCAL_MACHINE\System\CurrentControlSet\Services\EventLog\Application
 *                                          \<Our Service Name>
 * 
 * @exception com.ms.lang.RegKeyException - thrown if there was a problem dealing
 *          with these keys in the registry.
 */
    protected void removeRegistryKeys()throws com.ms.lang.RegKeyException{
            com.ms.lang.RegKey myKey = new com.ms.lang.RegKey(
                            com.ms.lang.RegKey.LOCALMACHINE_ROOT,
  "SYSTEM\\CurrentControlSet\\Services\\EventLog\\Application\\",
                            com.ms.lang.RegKey.KEYOPEN_WRITE);
            myKey.deleteSubKey(this.getServiceName());
    }
/**
 * 
 */
    protected void CmdDebugService(String[] args){
        System.out.println("Debugging "+getServiceDisplayName()+".");
        Advapi32.SetConsoleCtrlHandler(com.ms.dll.DllLib.addrOf(
                     com.ms.dll.Root.alloc(new ServiceControl())), true);
        // TODO : convert the java args to C command line args.
        serviceStart(0,0);
    }
/**
 * When a service receives a control request, the service's Handler function must 
 * call SetServiceStatus, even if the service's status did not change. A service can
 * also send the status at any time and by any thread of the service to notify 
 * the service control manager of status changes. Examples of such unsolicited 
 * status updates include: 
 * 
 *      Checkpoint updates that occur when the service is in transition from one 
 *      state to another (that is, SERVICE_START_PENDING). 
 * 
 *      Fatal error updates that occur when the service must stop due to a 
 *      recoverable error. 
 * 
 * A service can call this function only after it has called 
 * RegisterServiceCtrlHandler to get a service status handle.
 * 
 * This method is used by the ServiceMain and ServiceControl classes to report
 * status to the SCM, and can be used by any subclass of NTService for reporting 
 * status.
 * 
 * @param int dwCurrentState -  Indicates the current state of the service. One of 
 *          the following values is specified: 
 * 
 *          Value                       Meaning
 * 
 *          SERVICE_STOPPED             The service is not running. 
 * 
 *          SERVICE_START_PENDING       The service is starting. 
 * 
 *          SERVICE_STOP_PENDING        The service is stopping. 
 * 
 *          SERVICE_RUNNING             The service is running. 
 * 
 *          SERVICE_CONTINUE_PENDING    The service continue is pending. 
 * 
 *          SERVICE_PAUSE_PENDING       The service pause is pending. 
 * 
 *          SERVICE_PAUSED              The service is paused. 
 * 
 * @param int dwWin32ExitCode - Specifies an Win32 error code that the service uses to report an error that 
 *      occurs when it is starting or stopping. To return an error code specific to 
 *      the service, the service must set this value to ERROR_SERVICE_SPECIFIC_ERROR 
 *      to indicate that the dwServiceSpecificExitCode member contains the error code.
 *      The service should set this value to NO_ERROR when it is running and on 
 *      normal termination. 
 * 
 * @param int dwWaitHint - Specifies an estimate of the amount of time, in 
 *          milliseconds, that the service expects a pending start, stop, pause, or 
 *          continue operation to take before the service makes its next call to the
 *          SetServiceStatus function with either an incremented dwCheckPoint value 
 *          or a change in dwCurrentState. If the amount of time specified by 
 *          dwWaitHint passes, and dwCheckPoint has not been incremented, or 
 *          dwCurrentState has not changed, the service control manager or service 
 *          control program can assume that an error has occurred.
 * 
 * @return boolean - indicates if the write to the SCM was successful.
 * 
 * @see sample.ntservice.Advapi32#SetServiceStatus
 * @see sample.ntservice.SERVICE_STATUS
 * @see sample.ntservice.SERVICE_STATUS_HANDLE
 */
	protected synchronized boolean ReportStatusToSCMgr(int dwCurrentState,
                                            int dwWin32ExitCode,
                                            int dwWaitHint){
        int dwCheckPoint=1;
        boolean fResult=true;
        
        if (dwCurrentState==com.ms.win32.win.SERVICE_START_PENDING)
            ssStatus.dwControlsAccepted=0;
        else 
            ssStatus.dwControlsAccepted=com.ms.win32.win.SERVICE_ACCEPT_STOP|
                                        com.ms.win32.win.SERVICE_ACCEPT_PAUSE_CONTINUE;
        ssStatus.dwCurrentState=dwCurrentState;
        ssStatus.dwWin32ExitCode=dwWin32ExitCode;
        ssStatus.dwWaitHint=dwWaitHint;
        if ((dwCurrentState==com.ms.win32.win.SERVICE_RUNNING)||
            (dwCurrentState == com.ms.win32.win.SERVICE_STOPPED))
            ssStatus.dwCheckPoint=0;
        else
            ssStatus.dwCheckPoint=dwCheckPoint++;
        
        //Report the status of the service to the service control manager.
        //
        if (!(fResult=Advapi32.SetServiceStatus(sshStatusHandle,ssStatus)))
            this.logEvent(null,new String[]{"SetServiceStatus failed "+
                          com.ms.dll.DllLib.getLastWin32Error()},
                          (short)com.ms.win32.win.EVENTLOG_ERROR_TYPE,
                          null,2000,3);
         return fResult;
	}
/**
 * The logEvent method writes an entry at the end of the event log. The entry is 
 * written to the end of the logfile for the source identified by the hEventLog 
 * parameter. The logEvent method adds the time, the entry's length, and the offsets
 * before storing the entry in the log. 
 * 
 * @param String pFormat - The Description to be used in the message
 * 
 * @param short messageType - Specifies the type of event being logged. This 
 *          parameter can be one of the following values: 
 * 
 *          Value                       Meaning 
 * 
 *		    EVENTLOG_ERROR_TYPE         Error event 
 *			EVENTLOG_WARNING_TYPE       Warning event 
 *			EVENTLOG_INFORMATION_TYPE   Information event 
 *			EVENTLOG_AUDIT_SUCCESS      Success Audit event 
 *			EVENTLOG_AUDIT_FAILURE      Failure Audit event 
 * 
 * 
 * @param byte[] data - Any data that should be contained with the event message, 
 *          that will be displayed as a list of bytes.
 */
	protected synchronized void logEvent(String machine,String[] pFormat, 
                                         short messageType,
                                         byte[] data,
                                         int eventID,
                                         int category){
        int dataLength=0;
        int pFormatLen=0;
        if (pFormat!=null)
            pFormatLen=pFormat.length;
        if (data!=null)
            dataLength=data.length;
    	int hEventSource = Advapi32.RegisterEventSource(machine,this.getServiceName());
		if (hEventSource !=0){
			// Write to event log.
			boolean b = Advapi32.ReportEventW(hEventSource, // handle to event source
								 messageType,      // type of message
								 (short)category, //Category
                                 eventID,        //Event ID
                                 null,     // Users SID (can be null)
                                 (short)pFormatLen, // number of strings in the String array
                                 dataLength, //length of the message
                                 pFormat,//array containing the message
								 data); //user byte array data
			Advapi32.DeregisterEventSource(hEventSource);
		}
	}
//*********************************************************************************
/**
 * Implementation of this class is used as a wrapper for a C function pointer, that
 * is passed in the RegisterServiceCtrlHandler function call in the ServiceMain
 * callback.  This class will be used to handle the requests of the SCM for 
 * Stopping/Pausing/Continuing the service.  Once the message has been recieved, 
 * this callback will invoke the users implementation of the abstract methods for
 * stop/pause/continue.
 * 
 * @see sample.ntservice.NTService.ServiceMain
 * @see com.ms.dll.Callback
 */
    class ServiceControl extends com.ms.dll.Callback{
/**
 * This method is called by the SCM whenever ControlService() is called on this 
 * service.
 * 
 * @param int dwCtrlCode - type of control requested.
 */
	public void callback(int dwCtrlCode){
		// Handle the requested control code.
        //
		switch(dwCtrlCode){
		case com.ms.win32.win.SERVICE_CONTROL_PAUSE:
			this$0.servicePause();
			this$0.ssStatus.dwCurrentState = com.ms.win32.win.SERVICE_PAUSED;
			break;
		case com.ms.win32.win.SERVICE_CONTROL_CONTINUE:
			serviceContinue();
			this$0.ssStatus.dwCurrentState = com.ms.win32.win.SERVICE_RUNNING;
			break;
		case com.ms.win32.win.SERVICE_CONTROL_STOP:
			this$0.ReportStatusToSCMgr(
                                com.ms.win32.win.SERVICE_STOP_PENDING,
                                com.ms.win32.win.NO_ERROR,
                                this$0.getServiceStopTimeout());
            this$0.serviceStop();
            synchronized (this$0.stopevt){
                this$0.stopevt.notify();
            }
            return;
		case com.ms.win32.win.SERVICE_CONTROL_INTERROGATE:
			// Proceed to set the status
			break;
            
        case com.ms.win32.win.SERVICE_CONTROL_SHUTDOWN:
            this$0.serviceShutdown();
            this.callback(com.ms.win32.win.SERVICE_CONTROL_STOP);
            break;
		default:
			System.out.println("ILLEGAL OPERATION CODE");
			return;
		}
        this$0.ReportStatusToSCMgr(this$0.ssStatus.dwCurrentState,
                                        com.ms.win32.win.NO_ERROR,
                                        0);
	}
}
//**********************************************************************************
/**
 * This class represents the function pointer that is required in the 
 * SERVICE_TABLE_ENTRY that is sent to the SCM indicating the main method for 
 * our service.  The SCM will call the callback method in this class when it wants
 * to start our service, and will use the ServiceControl callback specified by this
 * class when it wants to send control messages (such as stop/pause/continue) to
 * us.
 * 
 * @see com.ms.dll.Callback
 * @see sample.ntservice.NTService.ServiceControl
 */
    class ServiceMain extends com.ms.dll.Callback{
/**
 * Performs the actual initialization of the service.  This routine performs the 
 * service initialization and then calls the user defined ServiceStart() routine
 * to perform the majority of the work.
 * 
 * @param int dwArgc - number of command line arguments.
 * @param int lpszArgv - array of command line arguments.
 */
	public void callback(int dwArgc, int lpszArgv){
        // register our service control handler:
        //
        sshStatusHandle = 
					Advapi32.RegisterServiceCtrlHandler(
            this$0.getServiceName(),
			com.ms.dll.DllLib.addrOf(com.ms.dll.Root.alloc(new ServiceControl())));
     
        if (sshStatusHandle==null){
            emergencyShutdown("Register ServiceCtrlHandler error "+
							   com.ms.dll.DllLib.getLastWin32Error());
            return;
        }
        ssStatus.dwServiceType=com.ms.win32.win.SERVICE_WIN32_OWN_PROCESS;
        ssStatus.dwServiceSpecificExitCode=0;
        // Specify that this service will listen to stop/pause and continue events
        ssStatus.dwControlsAccepted=com.ms.win32.win.SERVICE_ACCEPT_PAUSE_CONTINUE|
                                    com.ms.win32.win.SERVICE_ACCEPT_STOP;
        // report the status to the service control manager.
        //
        if (!ReportStatusToSCMgr(
                       com.ms.win32.win.SERVICE_START_PENDING,
                       com.ms.win32.win.NO_ERROR,
                       this$0.getServiceStartTimeout())){
            emergencyShutdown("Unable to report status to SCM ");
            return;
        }
        // Call the users 'start' method to actually start the program we are
        // running as an NT Service. 
        serviceStart(dwArgc,lpszArgv);
        // Report to the SCM that our program is in fact running.  If this 
        // report is NOT sent in the time we specified in the getServiceTimeout
        // method the SCM will have already determined that there was a problem
        // and will display an error message to the user.
        if (!ReportStatusToSCMgr(
                     com.ms.win32.win.SERVICE_RUNNING,
                       com.ms.win32.win.NO_ERROR,
                       0)){
            //There was a problem reporting our state to the SCM, so we are 
            //shutting down.
            emergencyShutdown("Unable to report status to SCM ");
                return;
        }
        // Wait until the Control receives a stop request before sending the 
        // stopped message to the SCM.
        synchronized (stopevt){
            try{
                stopevt.wait();
            }
            catch (InterruptedException ie){}
        }
        // Stop event was detected, so send the stopped message to the SCM.
        if (!ReportStatusToSCMgr(
                       com.ms.win32.win.SERVICE_STOPPED,
                       com.ms.win32.win.NO_ERROR,
                       0))
                  logEvent(null,new String[]{
                                "Service stopped, but unable to report status to SCM "},
                          (short)com.ms.win32.win.EVENTLOG_ERROR_TYPE,
                          null,2000,3);
         }
    }
/**
 * If there are any problems encountered when running in our service, this 
 * method is called to do any work required to immediately shutdown our service.
 * 
 * @param String message - This message will be sent to the event log as an 
 *          error.
 */
    protected void emergencyShutdown(String message){
        logEvent(null,new String[]{
            message},
            (short)com.ms.win32.win.EVENTLOG_ERROR_TYPE,
            null,2000,3);
        //Perhaps this should be a call to the SCM telling it to shutdown our 
        // service as well.
        serviceStop();
     }
}