/*
 *
 * sample.decoder.AttributeObject.java
 *
 * (C) Copyright 1995 - 1999 Microsoft Corporation.  All rights reserved.
 *
*/
package sample.decoder;

import java.util.Vector;
import java.util.Hashtable;
import java.lang.reflect.Method;
/**
 * Basic Object for containing any 'attribute_info' structures from the java 
 * class byte code.  Using the attribute name as a suffix for a  
 * sample.decoder.<name>Attribute.class file which will decode the 
 * attribute in more detail, this class attempts to use reflection in 
 * locating this object.  If the object is not found, this object will decode
 * the bytes using the default interpretation (basically none) for the 
 * contents of this attribute.  Adding support for new attributes simply 
 * requires constructing a class with the name <new attribute>Attribute which
 * extends AttributeObject and is also in this package.
 *
 * The constructor in this method takes a Vector of the already discovered
 * Constant Pool objects, and uses this to retrieve the actual constant pool 
 * objects that each attribute may have indexs to.  This will allow users of
 * these objects, after they have been decoded, to deal with each Attribute 
 * without the dependency on the ConstantPool objects.  
 *
 * @see sample.decoder.ClassObject
 * @see sample.decoder.CodeAttribute
 * @see sample.decoder.ConstantValueAttribute
 * @see sample.decoder.ExceptionsAttribute
 * @see sample.decoder.LineNumberTableAttribute
 * @see sample.decoder.LocalVariableTableAttribute
 * @see com.me.tools.decode.SourceFileAttribute
 *
 * @date	April 28,1998
 * @version	1.2
*/
public class AttributeObject extends DecodeObject implements 
			DecodeDisplayInterface,java.io.Serializable{
/**
 * Denotes the source file attribute info structure.
*/
	public final String SOURCEFILE_ATTRIBUTE="SourceFile";
/**
 * Specifies the Constant Value Attribute structure.
*/
	public final String CONSTANTVALUE_ATTRIBUTE="ConstantValue";
/**
 * Indicates the Code Attribute structure.
*/
	public final String CODE_ATTRIBUTE="Code";
/**
 * Specifies the Exception Attribute structure,
*/                                                                                                                                                                                       
	public final String EXCEPTIONS_ATTRIBUTE="Exceptions";
/**
 * Denotes the Line Number Table Attribute structure.
*/
	public final String LINENUMBERTABLE_ATTRIBUTE="LineNumberTable";
/**
 * Indicates the Local Variable Table structure.
*/
	public final String LOCALVARIABLETABLE_ATTRIBUTE="LocalVariableTable";
/**
 * This is a temporary Vector of Constant pool objects used for retrieving
 * any Constant Pool objects that may be indexed by an Attribute object or
 * a subobject of an Attribute.
 *
 * @see sample.decoder.ConstantObject
*/
	private transient Vector constantPool = null;
/** 
 * Bytes that represent the attribute_name_index into the ConstantPool.  The
 * index will point to a Constant_Utf8 Object containing a String that 
 * specifies the type of object.
 *
 * @see #CODE_ATTRIBUTE
 * @see #EXCEPTIONS_ATTRIBUTE
*/
	protected byte[] attribute_name_index=new byte[2];
/** 
 * A ConstantUtf8Object which represents the String that was referenced by 
 * the bytes attribute_name_index.
 *
 * @see #attribute_name_index.
*/
	protected ConstantUtfObject name = null;
/**
 * This is the number of bytes that this attribute structure occupies in the
 * classfile.  It is necessary for moving onto the next attribute if the
 * actual decomposition of this attribute/or its type is not known.
*/
	protected byte[] attribute_length = new byte[4];
/**
 * Creates an instance of Attribute Object and initializes the constantPool
 * vector to the Vector of ConstantPool objects passed as a parameter.  Note
 * this if this Vector does not contain the correct Constant Pool objects or
 * is null, the decoding will not complete.
 *
 * @param Vector pool contains all of the decoded constant pool objects
 *
 * @see sample.decoder.ConstantPool
*/
public AttributeObject(Vector pool){
        constantPool=pool;	
}
/**
 * This empty constructor is required to allow 'stuffing' new instances of
 * attribute object as a default container for unknown attributes.
*/
public AttributeObject(){}
/**
 * After the initial decoding phase is complete, where all of the bytes have
 * been moved to their appropriate places, this method retrieves any 
 * referenced Constant Pool objects and stores the information with the 
 * object containing the refernece.
 *
 * @param Vector data - These are the ConstantObjects that were created from the 
 *						constant pool structures in the class file.
*/
public void link(Vector data){
	try{
		name=(ConstantUtfObject)data.elementAt(getAttributeNameIndex());
	}catch(Exception ii){
		System.out.println("AttributeObject: Exception linking name "+ii+
			" "+getAttributeNameIndex());
	}
}
/**
 * Default method for determining a display String which adequatly 
 * represents this object and its contents.  Subclasses of this object 
 * should override this method to provide more detailed information.
 *
 * @returns String representing this object.
*/
public String toString(){
	return getName();//a+b+c+d;
}
/**
 * This method is a member of the DecodeDisplayInterface and is used to 
 * return the object name and the attribute name.  Subclasses of this object
 * should over ride this method to provide more detailed information.
 *
 * @see sample.decoder.DecodeDisplayInterface
 *
 * @returns String name representing this object and its attribute type.
*/
public String getName(){
	return getAttributeName();
}
/**
 * Required by the DecodeDisplayInterface, returns any Vectors, or subobjects
 * which will need to be displayed as a subobject in a GUI.  The Hashtable is
 * indexed by a String that is to be used as a display name, the contents
 * will be either a Vector of Strings/Objects or a Decode Object.
 * 
 * @returns Hashtable of subobjects/structures to be displayed as seperate
 *				nodes in the GUI.
*/
public Hashtable getChildren(){
	return new Hashtable();
}
/**
 * Returns the String contained in the ConstantUtf8Object which represents
 * the type of attribute that this is.
 *
 * @returns String with attribute name.
*/
public String getAttributeName(){
		return name.toString();
}
/**
 * This is the length of the attribute and all of its subcomponents.  If the
 * actual decomposition of an attributes subcomponents is not available this
 * number is used to determine where the next attribute starts.
 *
 * @returns int representing the number of bytes this attribute takes up.
*/
public int getAttributeLength(){
	return Decode.getIntegerValue(attribute_length);
}
/**
 * The index into the ConstantPool which holds the ConstantUtf8Object 
 * representing the name of this attribute.
 *
 * @returns int index into the constant pool.
*/
public int getAttributeNameIndex(){
	return Decode.getIntegerValue(attribute_name_index);
}
/**
 * Creates a Vector of the speficied number of attributes from the stream it
 * is passed starting at the indicated index into the stream.
 *
 * @param byte[] classBytes - the bytes representing the class to decode.
 * @param int indexs -  this is the number of attributes expected.
 * @param int startByte - the index into the classBytes to start decoding.
 *
 * @returns Vector containing AttributeObjects or its subclasses.
*/
public Vector createObjects(byte[] classBytes,int indexs,int startByte)throws Exception{
	log("AttributeObject: Looking for "+indexs+" attributes.");
	Vector list = new Vector();
    for (int i=0;i<indexs;i++){
		Object newton=null;
        if (list.size()!=0)
			newton=nextObject(
				classBytes, ((DecodeObject)list.elementAt(i-1)).
                getByteStopIndex()+1);
        else
			newton=nextObject(classBytes, startByte);
        ((DecodeObject)newton).index=i+1;
        list.addElement(newton);
	}
	return list;
}
/**
 * Determines the next AttributeObject or its subclass in the given stream
 * by starting to look for an attribute at the given index.  This method
 * uses reflection in an attempt to locate a subclass of AttributeObject
 * which can decode the field for this attribute.  The specific class would
 * be in the same package, and have the name"<attributename>Attribute.class".
 * If the spefific Attribute class cannot be found, the default 
 * AttributeObject instance is returned.
 *
 * @param byte[] classBytes - the class as bytes to be decoded.
 * @param int index - the index into the byte stream to start looking.
 *
 * @returns Object - the AttributeObject for the next attribute in the stream.
*/
public Object nextObject(byte[] classBytes, int index)throws Exception{
	AttributeObject newFound = new AttributeObject();
    newFound.setByteStartIndex(index);
    newFound.attribute_name_index[0]=classBytes[index];
    newFound.attribute_name_index[1]=classBytes[index+1];
	newFound.addEntry("Attribute Name",index,2,constantPool.elementAt(Decode.getIntegerValue(
		newFound.attribute_name_index)).toString(),newFound.attribute_name_index);
	log("AttributeObject: index into the constant pool is "+
		Decode.getIntegerValue(newFound.attribute_name_index));
	String attributeType=((ConstantUtfObject)(
		constantPool.elementAt(Decode.getIntegerValue(newFound.attribute_name_index))
		)).toString();	
	try{
		String me = this.getClass().getName();
		String toGet = me.substring(0,me.lastIndexOf("."))+"."+attributeType+"Attribute";
		log ("AttributeObject: class to load is "+toGet);
		newFound= (AttributeObject)getSpecificInstance(toGet,classBytes, index,constantPool);
	}catch(Exception ii){
		log("AttributeObject: Exception getting specific info for "+attributeType+" "+ii);
		newFound =  (AttributeObject)getInstance(new AttributeObject(),
			classBytes, index,constantPool,true);
		System.out.println(attributeType+" "+
						   Decode.getByteString(newFound.contents));
	}
	newFound.link(constantPool);
	return newFound;
}
/**
 * Tries to load the specialized class for decoding the current attribute.
 * If this method fails in any way, the generic attribute decoding takes 
 * place.
 *
 * @param String cname - The constructed class name for the specific decode
 *				Object.
 * @param byte[] stream  - the byte stream representing the class that is
 *				being decoded.
 * @param int index - index into the byte stream where this attribute starts.
 * @param Vector constantPool - all constant pool objects which have been 
 *				decoded.
 *
 * @returns Object -  next Attribute Object from the stream.
*/
private Object getSpecificInstance(String cname, byte[] stream, int index,
								   Vector constantPool) throws Exception{
	Class c = Class.forName(cname);
	Object[] param = new Object[3];
	param[0]=stream;
	param[1]=new Integer(index+6);
	param[2]=constantPool;
	Method[] m = c.getDeclaredMethods();
	int met = 0;
	for (int i=0;i<m.length;i++)
		if (m[i].getName().equals("getInstance")) met=i;
	log("AttributeObject: "+cname+" method index is "+met+" name "+m[met].getName());
	AttributeObject attr =  (AttributeObject)m[met].invoke(c.newInstance(),param);
	return getInstance(attr,stream, index,constantPool,false);
}
/**
 * Generic method for writting debug information to the screen.  By having all
 * methods in this class send their log information here, there is a central
 * point of change for turning debuggin off, or sending it to a differnet 
 * location.
 *
 * @param String data - message to be displayed.
*/
public synchronized void log(String data){
//		System.out.println(data);
}
/**
 * Creates a default AttributeObject out of the byte stream starting from the
 * specified index.
 *
 * @param byte[] stream - bytes representing the class currently being decoded
 * @param int index - starting byte for this attribute in the stream.
 * @param Vector constantPool - All ConstantPool objects decoded.
*/
public static Object getInstance(AttributeObject xconst, 
								 byte[] stream, 
								 int index,
								 Vector constantPool,
								 boolean addContents){
        xconst.setByteStartIndex(index);
        xconst.attribute_name_index[0]=stream[index];
        xconst.attribute_name_index[1]=stream[index+1];
        xconst.addEntry("Attribute Name",index,2,
			constantPool.elementAt(Decode.getIntegerValue(xconst.attribute_name_index)).toString(),
			xconst.attribute_name_index);
		xconst.attribute_length[0]=stream[index+2];
        xconst.attribute_length[1]=stream[index+3];
        xconst.attribute_length[2]=stream[index+4];
        xconst.attribute_length[3]=stream[index+5];
		xconst.addEntry("Attribute Length",index+2,4,""+Decode.getIntegerValue(xconst.attribute_length),
			xconst.attribute_length);
		index+=6;
        xconst.contents=new byte[Decode.getIntegerValue(xconst.attribute_length)];
        for (int i=0;i<xconst.contents.length;i++)
                xconst.contents[i]=stream[index++];
		if (addContents)
			xconst.addEntry("Attribute Contents",index-1-xconst.contents.length,
				xconst.contents.length,constantPool.elementAt(Decode.getIntegerValue(xconst.attribute_name_index)).toString(),
				xconst.contents);
	    xconst.setByteStopIndex(index-1);
        return xconst;
}
}