/*
 *
 * sample.decoder.COM_Class_TypeAttribute.java
 *
 * (C) Copyright 1995 - 1999 Microsoft Corporation.  All rights reserved.
 *
 */
package sample.decoder;

import java.util.Vector;
import java.util.Hashtable;
/**
 * Attribute Name:  COM_Class_Type
 * Attribute Level: Class
 * Restrictions:
 *    If the class type is MCCTCT_JCW, the access flags for the
 *      class must be a subset of:
 *
 *          ACC_PUBLIC
 *          ACC_FINAL
 *          ACC_ABSTRACT
 *
 *    If the class type is MCCTCT_JCDW, the access flags for the
 *      class must be a subset of:
 *
 *          ACC_PUBLIC
 *          ACC_FINAL (required)
 *
 *    The class's superclass must be java.lang.Object.
 *
 *    If the class type is MCCTCT_JCDW, the ACC_FINAL access flag must
 *    be present, and all fields defined by the class must have
 *    a COM_MapsTo. 
 *
 *
 *
 * Attribute elements:
 * 
 *     WORD Flags
 *         CCTF_NODYNAMICIFCCASTS:
 *            This bit is legal only if the ClassType is MCCTCT_JCW.
 *
 *            If on, instances of this class can only be legally
 *            casts to interfaces explicitly implemented by either
 *            the instance's class or one of its ancestors.
 *
 *            If off, instances of this class can be cast at runtime
 *            to any interface which contains a methodpool item,
 *            provided that the underlying COM object returns a valid
 *            interface for the IID of the interface.
 *
 *
 *         CCTF_CHARSIZEMASK
 *
 *            These bits must be zero unless ClassType is MCCTCT_JCDW.
 *
 *            Otherwise, these bits indicate the treatment of certain
 *            string and character array types:
 *
 *              CCTF_CHARSIZEDUAL       -- dual mode
 *              CCTF_CHARSIZEANSI       -- ansi
 *              CCTF_CHARSIZEUNICODE    -- unicode
 *
 *
 *     WORD ClassType
 *         The only legal values are MCCTCT_JCW and MCCTCT_JCDW.
 *
 *         The actual type of a class is determined as follows:
 *
 *           If no ancestor class (including the class itself) has a ClassType
 *           attribute, the type is "pure".
 *
 *           If the class has no ClassType attribute but a single ancestor
 *           (which must be a direct descendant of java.lang.Object) has a
 *           ClassType attribute JCW, the type is "Derived from JCW."
 *
 *            If the class has a ClassType, of type "MCCTCT_JCW", the class
 *             type is "JCW."
 *
 *           If the class has a ClassType, of type "MCCTCT_JCDW", the class
 *           type is "JCDW."
 *
 *           All other situations are illegal.
 *
 *
 *     WORD CLSIDIndex
 *
 *          If ClassType is not MCCTCT_JCW, the contents of the field
 *          must 0xffff.
 *
 *          If the ClassType is MCCTCT_JCW, and the CLSIDIndex is
 *          not 0xffff, it must indicate a valid index into the
 *          GuidPool. The GUID corresponding to this index will be
 *          used as the CLSID of the COM object to load when
 *          an instance of this class is created using the "new"
 *          operator.
 *========================================================================
 * #define CCTF_NODYNAMICIFCCASTS 0x0001
 *
 * #define CCTF_CHARSIZEMASK      0x0006
 * #define CCTF_CHARSIZEDUAL      0x0000    
 * #define CCTF_CHARSIZEANSI      0x0002
 * #define CCTF_CHARSIZEUNICODE   0x0004
 *
 * #define MCCTCT_JCW      ((WORD)1)
 * #define MCCTCT_JCDW     ((WORD)2)
 *
 * typedef struct {
 *    WORD        Flags;          // CCTF_* value.
 *    WORD        Classtype;      // MCCTCT_JCW or MCCTCT_JCDW
 *    union {
 *        WORD        CLSIDIndex;     // Optional index to CLSID (-1 for none)
 *        struct {
 *            BYTE    packing;    //1,2,4 or 8 (0xff defaults to 4 for backward compatibility)
 *            BYTE    reserved1;  // must set to 0xff
 *        } jcdw;
 *    };
 * } MCCT_Attr;
 * 
 * SCOPE : Class
 */
public class COM_Class_TypeAttribute extends AttributeObject
    implements DecodeDisplayInterface, java.io.Serializable{
    
	private final static int    CCTF_NODYNAMICIFCCASTS = 0x0001;

    private final static int    CCTF_CHARSIZEMASK = 0x0006;
    private final static int    CCTF_CHARSIZEDUAL = 0x0000;
    private final static int    CCTF_CHARSIZEANSI = 0x0002;
    private final static int    CCTF_CHARSIZEUNICODE = 0x0004;

    private final static int    MCCTCT_JCW = 1;
    private final static int    MCCTCT_JCDW = 2;

    private int     m_nFlags;
    private int     m_nClasstype;
    private int     m_nJCDWPacking;
    private String  m_strCLSIDForJCW;

    // used only before resolution.  index into guid pool.
    private int     m_nCLSIDForJCWIndex;

/**
 * After the initial decoding phase is complete, where all of the bytes have
 * been moved to their appropriate places, this method retrieves any 
 * referenced Constant Pool objects and stores the information with the 
 * object containing the refernece.
 *
 * @param Vector data - These are the ConstantObjects that were created from the 
 *						constant pool structures in the class file.
 */
    public void link( Vector data ){
        super.link( data );
        if( m_nClasstype == MCCTCT_JCW )
            m_strCLSIDForJCW = "";
	}
/**
 * This method is a member of the DecodeDisplayInterface and is used to 
 * return the object name and the attribute name.  Subclasses of this object
 * should over ride this method to provide more detailed information.
 *
 * @see sample.decoder.DecodeDisplayInterface
 *
 * @returns String name representing this object and its attribute type.
 */
    public String getName(){
	    return "COM_Class_Type";
    }
/**
 * Creates a human readable String that represents the flags associated with this
 * attribute.
 * 
 * @return String - human readable string representing the flags. 
 */
	private String getFlagsString(){
		String      strFlags = "";
        if( (m_nFlags & CCTF_CHARSIZEANSI) != 0 )
            strFlags += "ansi ";
        else if( (m_nFlags & CCTF_CHARSIZEUNICODE) != 0 )
            strFlags += "unicode ";
        else
            strFlags += "auto ";
        if( (m_nFlags & CCTF_NODYNAMICIFCCASTS) != 0 )
            strFlags += "CCTF_NODYNAMICIFCCASTS ";
        return strFlags;
	}
/**
 * Default method for determining a display String which adequatly 
 * represents this object and its contents.  Subclasses of this object 
 * should override this method to provide more detailed information.
 *
 * @returns String representing this object.
 */
    public String toString(){
	    return getName();
    }
/**
 *       typedef struct {
 *           WORD        Flags;          // CCTF_* value.
 *           WORD        Classtype;      // MCCTCT_JCW or MCCTCT_JCDW
 *           union {
 *               WORD        CLSIDIndex;     // Optional index to CLSID (-1 for none)
 *               struct {
 *                   BYTE    packing;    //1,2,4 or 8 (0xff defaults to 4 for backward compatibility)
 *                   BYTE    reserved1;  // must set to 0xff
 *               } jcdw;
 *           };
 *       } MCCT_Attr;
 */
    int readAttribute( byte[] stream, int index, Vector constantPool ){
        // read first two words.
        m_nFlags = (stream[ index++ ] << 8) | stream[ index++ ];
		addEntry("Flags",index-2,2,getFlagsString(),
				 new byte[]{new Integer(m_nFlags).byteValue()});
		m_nClasstype = (stream[ index++ ] << 8) | stream[ index++ ];
		addEntry("Class Type",index-2,2,""+m_nClasstype,
				 new byte[]{new Integer(m_nClasstype).byteValue()});
        // read next word, 2 bytes.
        byte    b1 = stream[ index++ ],
                b2 = stream[ index++ ];
        if( m_nClasstype == MCCTCT_JCDW ){
            m_nJCDWPacking = b1;
			addEntry("JCDW Packing",index-2,1,""+m_nJCDWPacking
					 ,new byte[]{new Integer(m_nJCDWPacking).byteValue()});
        }else{
            m_nCLSIDForJCWIndex = (b1 << 8) | b2;
			addEntry("CLSID for JCW Index",index-2,2,""+m_nCLSIDForJCWIndex,
					 new byte[]{new Integer(m_nCLSIDForJCWIndex).byteValue()});
        }
        return index;
    }
/**
 * This is the workhorse for the object, it decodes an instance of itself starting at the
 * specified byte index into the stream provided.  This was created as a static method
 * rather than a constructor, as there may be times when this method returns a more
 * specialized object which is a subclass of itself.
 *
 * @param byte[] stream -  entire bytes of the class that is being decoded
 * @param int index - this is the index into the stream for the starting byte.
 * @param Vector constantPool - these are the constantPool objects, any index into the
 *								constant pool will be resolved by retrieving the object 
 *								at the required instance from this Vector.
 *
 * @returns Object - This is the decoded object which is a populated instance of this 
 *					class.
 */
    public static Object getInstance( byte[] stream, int index, Vector constantPool ){
        COM_Class_TypeAttribute attr = new COM_Class_TypeAttribute();
		  // do the real decoding for this attribute.
        index = attr.readAttribute( stream, index, constantPool );
        return attr;
    }
}