/*
 *
 * sample.decoder.ClassObject.java
 *
 * (C) Copyright 1995 - 1999 Microsoft Corporation.  All rights reserved.
 *
*/
package sample.decoder;

import java.io.*;
import java.util.Vector;
import java.util.Hashtable;
/**
 * This is the main container for decoded class files.  It holds the initial bytes  from
 * the class file stream, and contains all of the decoded objects.  It has constructors
 * for taking class file names, for specifying the file to be decoded, or a byte stream
 * that will be decoded.  The results of the decoding will populate this object, and
 * will throw an exception if it does not complete successfully.  This was done since the
 * class files are comprised of variable length tables, and if any once component within
 * the classfile has not been decoded successfully, there is no gaurantee that any other
 * portion will be decoded correctly.
 *
 * @see sample.decoder.ClassViewer
 * @see sample.decoder.AttributeObject
 * @see sample.decoder.ConstantObject
 * @see sample.decoder.MethodObject
 * @see sample.decoder.FieldObject
 *
 * @date	April 28,1998
 * @version	1.2
*/
public class ClassObject extends DecodeObject implements DecodeDisplayInterface, java.io.Serializable{
	public static final byte[] MAGIC = (new byte[]{
										  new Integer(0xCA).byteValue(),
										  new Integer(0xFE).byteValue(),
										  new Integer(0xBA).byteValue(),
										  new Integer(0xBE).byteValue(),
										});
/**
 * Holds the entire byte stream representing this class.
*/
	private byte[] classDefinition = null;
/**
 * Contains the 4 bytes representing the magic number for the class file.
*/
	private byte[] magic = new byte[4];
/** 
 * Contains the 2 bytes representing the minor version of java which generated this file
*/
	private byte[] minor_version = new byte[2];
/**
 * Holds the 2 bytes representing the major version of java which generated this file.
*/
	private byte[] major_version = new byte[2];
/**
 * Contains the 2 bytes representing the number of constant objects in the constant pool
*/
	private byte[] constant_pool_count = new byte[2];
/** 
 * Contains the 2 bytes representing the access flags for the class decoded.
*/
	private byte[] access_flags = new byte[2];
/**
 * Contains 2 bytes which represent the constant pool index where the constant class
 * object is that represents this class.
*/
	private byte[] this_class = new byte[2];
/** 
 * Contains 2 bytes which represent the constant pool index where the constant class
 * object is that represents this class's parent.
*/
	private byte[] super_class = new byte[2];
/**
 * These 2 bytes contain the number of interfaces implemented by this class.
*/
	private byte[] interfaces_count = new byte[2];	
/** 
 * This array of bytes are two byte indexs into the constant pool for class objects 
 * that represent the interfaces supported by the decoded class.
*/
	private byte[] interfaces = null;
/**
 * These 2 bytes indicate the number of fields in the class that is decoded.
*/
	private byte[] fields_count = new byte [2];
/**
 * 2 bytes representing the number of methods in the class file decoded.
*/
	private byte[] methods_count = new byte[2];
/**
 * This is a Vector of all MethodObjects generated from the .class file.
 * 
 * @see sample.decoder.MethodObject
*/
	private Vector methods = null;
/**
 * These 2 bytes represent the number of attributes the decoded class has.
*/
	private byte[] attributes_count = new byte[2];
/**
 * This is a Vector containing the decoded AttributeObjects (or a subclass of 
 * Attribute Object).
 *
 * @see sample.decoder.AttributeObject
*/
	private Vector attributes = null;
/**
 * This is a Vector of all the Constant Objects which have been decoded from the
 * specified class's constant pool bytes.
 *
 * @see sample.decoder.ConstantObject
*/
	private Vector constants = null;
/**
 * This is a Vector of all the field objects decoded from the specified class file.
 *
 * @see sample.decoder.FieldObject
*/
	private Vector fields = null;
/**
 * Stores the filename of the class if the class was loaded for decoding, and not
 * determined from a given stream of bytes
 */
	protected String fileName=null;
/**
 * This constructor attempts to load the spefcified file from the classname specified.
 * If full path name must be provided for the load operation to work.  Once the class is
 * loaded, decoding will begin.  Any errors with opening the class file or with decoding
 * the file will be thrown.
 *
 * @param String className - the classfile name and path to be loaded and decoded.
*/
public ClassObject(String className)throws Exception{
	fileName = className;
	classDefinition=getClassAsBytes(className);
	processClassBytes();
}
/**
 * This takes a byte stream that will be used as a class and decoded.  If this is not a
 * valid java class, exceptions will be thrown due to errors in the decoding.
 * 
 * @param byte[] byteClass - This is an array of bytes representing a valid java class.
*/
public ClassObject(byte[] byteClass)throws Exception{
	classDefinition=byteClass;
	processClassBytes();
}
/**
 * Returns the filename of the file that was used for decoding the class, or null
 * if the file was decoded from a given byte stream.
 * 
 * @return - the filename if it exists, null otherwise
 */
public String getFileName(){
	return fileName;
}
/**
 * This is a simple display method that can be used to show information about the 
 * decoded class to the command line.  It does not display the raw byte information
 * and may yield a large amount of information if huge class files are decoded.
 *
 * @returns String - Display String representing the ClassObject and decoded values.
*/
public String toString(){
        String a="\n     sample.decoder.ClassObject    \n\n";
        String b="Magic "+getMagic()+"\n";
        String c="Minor Version "+getMinorVersion()+"\n";
        String d="Major Version "+getMajorVersion()+"\n";
        String e="Constant Pool Count "+getConstantPoolCount()+"\n";
        String f="Access Flags "+getAccessFlags()+"\n";
        String g="This Class "+getThisClass()+"\n";
        String h="Super Class "+getSuperClass()+"\n";
        String i="Interfaces Count "+getInterfacesCount()+"\n";
        String j="Fields Count "+getFieldsCount()+"\n";
        String k="Methods Count "+getMethodsCount()+"\n";
        String l="Attributes Count "+getAttributesCount()+"\n";
		String m="Fields "+fields+"\n";
		String n="Methods "+methods+"\n";
		String o="Attributes "+attributes+"\n";
        return a+b+c+d+e+f+g+h+i+j+k+l+m+n+o;
}
/**
 * Returns the entire contstant pool
 * 
 * @return Vector - contains all constantPool elements.
 */
public Vector getConstantPool(){
	return this.constants;
}
/**
 * Determines the name of this class, as told by the Constant Class Object containing the
 * information.
 * 
 * @returns String - represents the runtime object name/type for the decoded class.
*/
public String getName(){
	return getThisClass();
}
/**
 * Required by the DecodeDisplayInterface, returns any Vectors, or subobjects
 * which will need to be displayed as a subobject in a GUI.  The Hashtable is
 * indexed by a String that is to be used as a display name, the contents
 * will be either a Vector of Strings/Objects or a Decode Object.
 *
 * @returns Hashtable of subobjects/structures to be displayed as seperate
 *				nodes in the GUI.
*/
public Hashtable getChildren(){
	Hashtable val = new Hashtable();
	val.put("ConstantPool", getConstants());
	val.put("Fields",getFields());
	val.put("Methods",getMethods());
	val.put("Attributes",getAttributes());
	return val;
}
/**
 * Returns an integer which represents the 4 byte value for the class magic number.
 *
 * @returns int - integer representation of 4 byte magic number.
*/
public int getMagic(){
        return Decode.getIntegerValue(magic);
}
/**
 * Retrieves the integer value of the two byte minor version number from the decoded 
 * class.
 *
 * @returns int - int value of 2 bytes
*/
public int getMinorVersion(){
        return Decode.getIntegerValue(minor_version);
}
/**
 * Retrieves the major version number as an integer from the decoded class file.
 *
 * @returns int - int representing the 2 bytes for major_version
*/
public int getMajorVersion(){
        return Decode.getIntegerValue(major_version);
}
/**
 * This is the number of constant pool structures in the constant pool.
 *
 * @returns int - represents the integer value of the 2 bytes for constant pool count
*/
public int getConstantPoolCount(){
        return Decode.getIntegerValue(constant_pool_count);
}
/**
 * Determines the integer equivalent of the two bytes representing the access flags for
 * the decoded class.
 *
 * @returns int - integer representation of the 2 bytes from the class file.
*/
public int getAccessFlag(){
	return Decode.getIntegerValue(access_flags);
}
/**
 * Analyzes the access flags for this class to determine if we are an interface.
 * 
 * @return boolean - true if we are an interface, false otherwise.
 */
public boolean isInterface(){
	if ((getAccessFlag()&Decode.pow(2,9))==Decode.pow(2,9))
		return true;
	else 
		return false;
}
/**
 * Constructs a Vector of Strings which represent the access flags for this class. Access
 * Strings could be:
 *
 *	ACC_PUBLIC			ACC_FINAL			ACC_SYNCHRONIZED		ACC_INTERFACE
 *	ACC_ABSTRACT
 *
 * @returns Vector of the above mentioned Strings which apply to this class.
*/
public Vector getAccessFlags(){
	Vector result = new Vector();
	if ((getAccessFlag()&Decode.pow(2,0))==Decode.pow(2,0))
		result.addElement("public");
	if ((getAccessFlag()&Decode.pow(2,4))==Decode.pow(2,4))
		result.addElement("final");
	if ((getAccessFlag()&Decode.pow(2,5))==Decode.pow(2,5))
		result.addElement("synchronized");
	if (isInterface())
		result.addElement("interface");
	if ((getAccessFlag()&Decode.pow(2,10))==Decode.pow(2,10))
		result.addElement("abstract");
	return result;	               
}
/**
 * This method creates a 'source like' String that describes this class object.
 * The String will contain the name of the class, imports, and extensions.
 * 
 * @return String - String describing this class using source syntax.
 */
public String getDescription(){
	String result = "";
	Vector flags = getAccessFlags();
	Vector interf = getInterfaces();
	for (int i=0;i<flags.size();i++)
		result+=(String)flags.elementAt(i)+" ";
	result+="class ";
	result+=getName()+" ";
	if (getSuperClass()!=null)
		result+="extends "+getSuperClass()+" ";
	if (interf.size()>0){
		result+="implements ";
		for (int i=0;i<interf.size();i++)
			result+=((ConstantClassObject)interf.elementAt(i)).
				getClassName()+" ";
	}
	return result;
}
/**
 * Determines which ranges of Strings in the description are actually the names of
 * other java Objects.  These character ranges can be used to mark the original 
 * Description String as hyperlinks or change the color when there are java objects
 * present.
 * 
 * @return Vector - contains java.awt.Point objects which hold the start and end
 *					position of all the names of java objects in the Description 
 *					String.
 */
	public Vector getObjectLinks(){
		Vector points = new Vector();
		String data = getDescription();
		if (getSuperClass()!=null)
		points.addElement(new java.awt.Point(data.indexOf(getSuperClass()),
				data.indexOf(getSuperClass())+getSuperClass().length()));
		for (int i=0;i<getInterfaces().size();i++){
			String da =((ConstantClassObject)getInterfaces().elementAt(i)).
					getClassName()+" ";
			points.addElement(new java.awt.Point(data.indexOf(da),
										data.indexOf(da)+da.length()-1));
		}
		return points;
}
/**
 * This is the ConstantClassObject from the constant pool which specifies the class that
 * has been decoded.  It is the constant_class_info structure indicated in the 
 * this_class index.
 *
 * @returns ConstantClassObject - represents the clas which has been decoded.
*/
public String getThisClass(){
        return ((ConstantClassObject)constants.
                elementAt(Decode.getIntegerValue(this_class))).getClassName();
}
/**
 * The ConstantClassObject that represents the super class of the decoded object.  It is
 * the constant_class_info structure that was indicated in the super_class index.
 *
 * @returns ConstantClassObject - representing the super class of this object.
*/
public String getSuperClass(){
		if (Decode.getIntegerValue(super_class)==0)
			return null;//Handles decoding java.lang.Object
		else
			return ((ConstantClassObject)constants.
                elementAt(Decode.getIntegerValue(super_class))).getClassName();
}
/**
 * This is the number of interfaces that the decoded class file supports.  It is 
 * calculated from 2 bytes which specify the number from the .class file 
 *
 * @returns int - number of interfaces implemented
*/
public int getInterfacesCount(){
        return Decode.getIntegerValue(interfaces_count);
}
/**
 * These are the ConstantClassObjects which represent the interfaces that are implemented
 * by the decoded class.  They have been retrieved from the constant pool by using the
 * 2 byte indexs into the constantpool specified in the interfaces byte array.
 *
 * @returns Vector - contains ConstantClassObjects for the interfaces implemented.
*/
public Vector getInterfaces(){
	log("interface length "+interfaces.length);
        Vector inter = new Vector();
        for (int i=0;i<interfaces.length;){
                byte[] intnum = new byte[2];
                intnum[0]=interfaces[i];
                intnum[1]=interfaces[i+1];
                inter.addElement(constants.elementAt(
                        Decode.getIntegerValue(intnum)));
				i+=2;
        }
        return inter;
}
/**
 * These are the ConstantObjects and subclasses of ConstantObject that represent all of 
 * the constant pool constants.
 *
 * @returns Vector - contains ConstantPool ConstantObjects and ConstantPool subclasses.
*/
public Vector getConstants(){
	return constants;
}
/**
 * This is the number of fields indside the decoded class file.
 *
 * @returns int - calculated fro the 2 bytes in the class file.
*/
public int getFieldsCount(){
        return Decode.getIntegerValue(fields_count);
}
/**
 * These are the decoded field objects which represent the fields inside the decoded 
 * class.
 *
 * @see sample.decoder.FieldObject
 *
 * @returns Vector - contains FieldObjects
*/
public Vector getFields(){
        return fields;
}
/**
 * This is the number of methods within the decoded class file.
 *
 * @returns int - this is the 2 byte integer value from the class file.
*/
public int getMethodsCount(){
        return Decode.getIntegerValue(methods_count);
}
/**
 * These are the MethodObjects that have been constructed from decoding the method
 * Structures in the .class file.  
 *
 * @see sample.decoder.MethodObject
 *
 * @returns Vector - contains MethodObjects
*/
public Vector getMethods(){
        return methods;
}
/**
 * This is the number of attributes that this class has.  This is not the total number
 * of attribute structures within the class file.
 *
 * @returns int - the number of attributes associated with the class file itself.
*/
public int getAttributesCount(){
        return Decode.getIntegerValue(attributes_count);
}
/**
 * These are the AttributeObjects associated with the class file itself.  These Objects
 * may also be subclasses of AttributeObject.
 *
 * @see sample.decoder.AttributeObject
 *
 * @returns Vector - containing AttributeObjects and AttributeObject subclasses.
*/
public Vector getAttributes(){
        return attributes;
}
/**
 *  This loads the class file from the path+filename supplied and returns the file as a
 * byte array.  Any exceptions encountered during this process are thrown to the calling
 * class.
 *
 * @param String className - this is the path+filename of the class to load.
 *
 * @returns byte[] - byte representation of the file that was loaded.
*/
public byte[] getClassAsBytes(String className) throws Exception{        
        FileInputStream in = new FileInputStream(className);
	boolean done=false;
	Vector byteList = new Vector(10000);
	int nextByte=0;
	int count=0;
	while(nextByte!=-1){
			count++;
                nextByte=in.read();
                if (nextByte!=-1) byteList.addElement(new Integer(nextByte));
	}
	byte[] classDef = new byte[byteList.size()];
	for (int i=0;i<byteList.size();i++)
                classDef[i]=((Integer)byteList.elementAt(i)).byteValue();
	return classDef;
}
/**
 *  This is the inital and main portion of the class decoding.  It populates some of the
 * byte[] fields in this class with the appropriate bytes and calls the methodObject, 
 * Attributeobject, ConstantObject, FieldObject classes to decode their portions.
 *
 * @see sample.decoder.AttributeObject
 * @see sample.decoder.ConstantObject
 * @see sample.decoder.FieldObject
 * @see sample.decoder.MethodObject
*/
public void processClassBytes()throws Exception{
	magic[0]=classDefinition[0];
	magic[1]=classDefinition[1];
	magic[2]=classDefinition[2];
	magic[3]=classDefinition[3];
	if (Decode.getIntegerValue(magic)!=Decode.getIntegerValue(MAGIC))
		throw new Exception("Not a Valid java .class file! \n magic# is "+Decode.getByteString(magic)+
			" should be "+Decode.getByteString(MAGIC));
	this.setByteStartIndex(0);
	this.setByteStopIndex(classDefinition.length);
	addEntry("Magic Number",0,4,Decode.getByteString(magic),magic);
	minor_version[0]=classDefinition[4];
	minor_version[1]=classDefinition[5];
	addEntry("Minor Version",4,2,""+Decode.getIntegerValue(minor_version),
			 minor_version);
	major_version[0]= classDefinition[6];
	major_version[1]=classDefinition[7];
	addEntry("Major Version",6,2,""+Decode.getIntegerValue(major_version),
			 major_version);
	constant_pool_count[0]=classDefinition[8];
	constant_pool_count[1]=classDefinition[9];
	addEntry("Constant Pool Count",8,2,""+Decode.getIntegerValue(constant_pool_count),
			 constant_pool_count);	
    log("ClassObject: about to decode constants...");
	ConstantObject cc = new ConstantObject();
    constants=cc.createObjects(classDefinition,
    Decode.getIntegerValue(constant_pool_count),9);
	log("ClassObject: got vector");
    int lastByte = ((DecodeObject)constants.lastElement()).getByteStopIndex();
	log("ClassObject: done decoding constants");
	access_flags[0]=classDefinition[lastByte+1];
    access_flags[1]=classDefinition[lastByte+2];
	addEntry("Access Flags",lastByte+1,2,""+this.getAccessFlagsString(),
			 access_flags);
    this_class[0]=classDefinition[lastByte+3];
    this_class[1]=classDefinition[lastByte+4];
	addEntry("This Class",lastByte+3,2,
		constants.elementAt(Decode.getIntegerValue(this_class)).toString(),
		this_class);
    super_class[0]=classDefinition[lastByte+5];
    super_class[1]=classDefinition[lastByte+6];
	if (Decode.getIntegerValue(super_class)!=0)
	addEntry("Super Class",lastByte+5,2,
		constants.elementAt(Decode.getIntegerValue(super_class)).toString(),
		super_class);
	else
		addEntry("Super Class",lastByte+5,2,"<none>",super_class);
    interfaces_count[0]=classDefinition[lastByte+7];
	interfaces_count[1]=classDefinition[lastByte+8];
	addEntry("Interfaces Count",lastByte+7,2,""+Decode.getIntegerValue(interfaces_count),
			 interfaces_count);	
	lastByte+=9;
	interfaces = new byte[Decode.getIntegerValue(interfaces_count)*2];
	for (int counter =0;counter<interfaces.length;counter++)
		interfaces[counter]=classDefinition[lastByte++];
	fields_count[0]=classDefinition[lastByte++];
	fields_count[1]=classDefinition[lastByte++];
	addEntry("Fields Count",lastByte-2,2,""+Decode.getIntegerValue(fields_count),
			 fields_count);	
	log("ClassObject: about to decode fields...");
	FieldObject ff = new FieldObject(constants);
	fields=ff.createObjects(classDefinition,
		Decode.getIntegerValue(fields_count),lastByte);
	if (fields.size()>0)
		lastByte=((DecodeObject)fields.lastElement()).getByteStopIndex()+1;
	methods_count[0]=classDefinition[lastByte++];
	methods_count[1]=classDefinition[lastByte++];
	addEntry("Methods Count",lastByte-2,2,""+Decode.getIntegerValue(methods_count),
			 methods_count);	
	log("ClassObject: about to decode methods...");
	MethodObject mm = new MethodObject(constants);
	methods=mm.createObjects(classDefinition,
	Decode.getIntegerValue(methods_count),lastByte);
	if (methods.size()>0)
		lastByte=((DecodeObject)methods.lastElement()).getByteStopIndex()+1;
	attributes_count[0]=classDefinition[lastByte++];
	attributes_count[1]=classDefinition[lastByte++];
	addEntry("Attributes Count",lastByte-2,2,""+Decode.getIntegerValue(attributes_count),
			 attributes_count);	
	log("ClassObject: about to decode attributes...");
	AttributeObject aa = new AttributeObject(constants);
	attributes=aa.createObjects(classDefinition,
	Decode.getIntegerValue(attributes_count),lastByte);
	log("ClassObject: decode complete...");
}
/**
 * Generic method for writting debug information to the screen.  By having all
 * methods in this class send their log information here, there is a central
 * point of change for turning debuggin off, or sending it to a differnet 
 * location.
 *
 * @param String data - message to be displayed.
*/
public synchronized void log(String val){
//	System.out.println(val);
}
/**
 * This is a simple main method to enable this class to be used from the command line.
 * It takes a path+filename string from the command line and uses it as a parameter to 
 * the ClassObject constructor.  The result of the decoding is displayed by calling the
 * toString method in the ClassObject.
 *
 * @param String - path+filename of the java class file to decode.
*/
public static void main(String args[]){
	ClassObject c = null;
	try{
        c = new ClassObject(args[0]);
        c.processClassBytes();
        System.out.println(c);
	}catch(Exception ii){
		System.out.println("ClassObject: Exception "+ii);
	}
}
}