/*
 *
 * sample.decoder.ClasspathLoader0.java
 *
 * (C) Copyright 1995 - 1999 Microsoft Corporation.  All rights reserved.
 *
*/
package sample.decoder;

import java.util.Vector;
import java.util.zip.ZipEntry;
/**
 * This Class is used to decode classes by reading them from the classpath.  It takes
 * a String representing the class, and returns the associated byte codes for it, if
 * they are available.  All methods in this class are related to dealing with classes
 * from the classpath.
 *
 * @date	April 28,1998
 * @version	1.1
*/
public class ClasspathLoader0 implements ClasspathLoader,java.io.Serializable{
	/**
	 * Conains the classname (java/lang/String) as a key, and the absolute path to
	 * the associated .class file as the data.  All classes found from directories
	 * in the classpath are stored here.
	 */
	private java.util.Hashtable zips = new java.util.Hashtable();
	/**
	 * The .class files found inside all files in the classpath are stored in this
	 * table.  The class name (java/lang/String) is used as a key, with the 
	 * absolute path to the associated zip file stored as the data.
	 */
	private java.util.Hashtable dirs = new java.util.Hashtable();
															   
/**
 * Basic constructor that initializes this object by calling makeClassList to 
 * comb the classpath and cache the lcoation of every class.
 * 
 * @see sample.decoder.ClasspathLoader0#makeClassList
 */
	public ClasspathLoader0(){
			this.makeClassList();
	}
/**
 * Reads the specified class file from the classpath and returns the
 * bytes.  The Class name must be specified like this:
 * 
 *				java.lang.String = java/lang/String.class
 * 
 * @param String filename - Name of the class to load.
 * 
 * @return byte[] - bytes that represent this class, or null if couldn't be loaded.
 */
	public byte[] getClassFromPath(String filename){
		if (filename==null) return null;
		filename=filename.replace('.','/');
		try{
		if (this.dirs.get(filename)!=null)
			return dirBytes((String)dirs.get(filename));
		else if (this.zips.get(filename)!=null)
			return zipBytes((String)zips.get(filename),filename);
		}catch(Exception ii){
			System.out.println("Exception getting "+filename+" from path "+ii);
			return null;
		}
		return null;
	}
/**
 * Reads the bytes corresponding to the specified absolute file name.
 * 
 * @param String fname - Absolute path name of the file to read.
 * 
 * @return byte[] - the bytes associated with the specified file.
 * 
 * @exception Exception - This exception may arise from any of the file manipulation
 *						activites required to locate and load the specified file.
 */
	private byte[] dirBytes(String fname)throws Exception{
		System.out.println("Reading from file "+fname);
		java.io.File f = new java.io.File(fname);
		java.io.InputStream is = new java.io.FileInputStream(fname);
		return readStream(is);
	}
/**
 * Reads the specified class file from the indicated zip file.  It is assumed that
 * the zip file is identified by its absolute path, and that the class name will
 * appear like (java/lang/String), so that locating it in the zip file only 
 * requires looking for a ZipEntry that has a ".class" appended.
 * 
 * @param String fname - absolute path name for the zip file.
 * @param String className - name of the class to load.
 * 
 * @return byte[] - bytes that define the specified class.
 * 
 * @exception Exception - any exception that may arise in locating, reading or 
 *						reading the specified file from the zip.
 */
	private byte[] zipBytes(String fname, String className)throws Exception{
		java.util.zip.ZipInputStream zi = new java.util.zip.ZipInputStream(
											new java.io.FileInputStream(fname));
		boolean done=false;
		int off=0;
		while (!done){
			byte[] hold = null;
			byte[] temp = new byte[0];
			ZipEntry ze = zi.getNextEntry();
			if (ze==null) 
				done=true;
			else{
					byte[] buf = new byte[1024];
					int len;
					while((len=zi.read(buf,off,1024))>=0){
						hold = new byte[len+temp.length];
						for (int i=0;i<temp.length;i++)
							hold[i]=temp[i];
						for (int i=0;i<len;i++)
							hold[i+temp.length]=buf[i];
						temp=hold;
					}//end of while
					if (ze.getName().equals(className+".class"))
						return hold;
			zi.closeEntry();
			}//end of else
		}
		return null;
	}
/**
 * Reads all bytes from the specified stream , returning the contents as a byte array.
 * 
 * @param java.io.InputStream is - Stream to read the bytes from.
 * 
 * @return byte[] - bytes associated with the stream.
 * 
 * @exception Exception - any Exception associated with reading the stream
 */
	private byte[] readStream(java.io.InputStream is)throws Exception{
		byte[] last = new byte[0];
		byte[] buf = new byte[512];
		byte[] total=null;
		int count=0;
		boolean done=false;
		while(!done){
			count = is.read(buf);
			total = new byte[last.length+count];
			for (int i=0;i<last.length;i++)
				total[i]=last[i];
			for (int i=0;i<count;i++)
				total[last.length+i]=buf[i];
			last=total;
			if (count<512)
				done=true;
		}
		if (last.length>0)
			return last;
		else
			return null;
	}
/**
 * This is a recursive method for mapping all of the files that are considered to be
 * java classes into our dirs table.  By maintaining a list of all classes, we can 
 * greatly reduce the amount of time taken to retrieve the bytes for a specific class
 * and can easily deduce when the required class is not available.
 * 
 * @param java.io.File dir - The directory to map.
 * @param String base - The base directory, used when deducing which directories are
 *						package names.
 */
	private void mapDirectory(java.io.File dir,String base){
		String[] files = dir.list();
		for (int i=0;i<files.length;i++){
			java.io.File f = new java.io.File(dir.getPath(),files[i]);
			if (f.isDirectory())
				mapDirectory(f,base);
			else{
				if (f.getName().indexOf(".class")!=-1){// The file is a class file
					String prefix = f.getAbsolutePath().substring(
								base.length(),f.getAbsolutePath().lastIndexOf(".class")
								);
					prefix=prefix.replace('\\','/');
					this.dirs.put(prefix,f.getAbsolutePath());
					//Store the class, and its absolute path in table.
				}
			}
		}//end of for loop
	}
/**
 * Reads the specified zip file, adding all of its contents that are java class files
 * to the zips table.
 * 
 * @param java.io.File f - The zip file to be mapped. 
 */
	private void mapFile(java.io.File f)throws Exception{
		java.util.zip.ZipFile z = new java.util.zip.ZipFile(f.getAbsolutePath());
		for (java.util.Enumeration i=z.entries();i.hasMoreElements();){
			java.util.zip.ZipEntry ze = (java.util.zip.ZipEntry)i.nextElement();	
			if (ze.getName().indexOf(".class")!=-1){
				String prefix = ze.getName().substring(0,
											ze.getName().indexOf(".class"));
				zips.put(prefix,f.getAbsolutePath());
			}//else
		}
	}
/**
 * Reads the classpath, mapping the directories and zip files.
 * 
 * @exception Exception - this may result when trying to access an invalid entry in 
 *					the classpath.
 */
	private void makeClassList(){
		String path =System.getProperty("java.class.path");	
		java.util.StringTokenizer t = new java.util.StringTokenizer(path,";");
		for (java.util.Enumeration i=t;i.hasMoreElements();){
			java.io.File f = new java.io.File((String)i.nextElement());	
			try{
			if (f.isDirectory())
				mapDirectory(f,f.getPath());
			else 
				mapFile(f);
			}catch(Exception ii){
			//	System.out.println("Exception mapping "+f);
			}
		}
	}
}