/*
 *
 * sample.decoder.DecodeObject.java
 *
 * (C) Copyright 1995 - 1999 Microsoft Corporation.  All rights reserved.
 *
*/
package sample.decoder;

import java.util.Vector;
/**
 * This is the parent of all objects used in the decoding process.  It contains basic
 * attributes (such as start and stop byte index) and a default container for bytes.
 * There are some default implementation of methods here, however many of the work 
 * method should be overridden for accomplishing specific decoding capabilities as
 * required.
 *
 * @date	April 28,1998
 * @version	1.1
*/
public abstract class DecodeObject implements java.io.Serializable{
/**
 * Contains all of the 'Entries' for the object, by holding all instances of 
 * Entry objects associated with an objects data.  Each set of information decoded
 * will be stored by a DecodeObject calling the addEntry method. 
 * 
 * @see sample.decoder.DecodeObject#addEntry
 */
	private Vector entri = new Vector();
/**
 * Tag is the integer number which indicates what type of constant structure the object
 * represents.
 *
 * @see sample.decoder.ConstantObject
*/
	protected int tag = 0;
/**
 * Index is the current index into the byte stream which essentially represents the 
 * location of the next byte to be processed.
*/
	protected int index = -1;
/**
 * The byte start index is the index into the class file byte stream where this object
 * started.
*/
	private int byteStartIndex = -1;
/**
 * Byte Stop Index is the last byte in the class file byte stream which is used by this 
 * object.
*/
	private int byteStopIndex=-1;
/**
 * This is a default container for holding bytes that are known to be part of this object
 * but have not been decoded into specific attributes or objects.
*/
	protected byte[] contents;
/**
 * Accessor method for setting the byte stop index value.
 *
 * @param int i - integer representing the index into the class byte[] where the last 
 *						byte defining this object resides.
*/
public void setByteStopIndex(int i){
	byteStopIndex=i;
}
/**
 * Accessor method for setting the byte start index value.
 *
 * @param int i - integer representing the index into the class byte[] where the first 
 *						byte defining this object resides.
*/
public void setByteStartIndex(int i){
        byteStartIndex=i;
}
/**
 * Accessor method for retrieving the byte start index value.
 *
 * @param int i - integer representing the index into the class byte[] where the first 
 *						byte defining this object resides.
*/
public int getByteStartIndex(){
        return byteStartIndex;
}
/**
 * Accessor method for retrieving the byte last index value.
 *
 * @param int i - integer representing the index into the class byte[] where the last 
 *						byte defining this object resides.
*/
public int getByteStopIndex(){
	return byteStopIndex;
}
/**
 * Used for retrieving a table name for the contents of the given decoded object.
 * 
 * @return String - Description of class type, identifier, and total byte length.
 */
public String getTitle(){
	 String className = this.getClass().getName();
	 String type = "";
	 if (className.indexOf("Attribute")!=-1) type="Attribute";
	 if (className.indexOf("Method")!=-1) type="Method";
	 if (className.indexOf("Field")!=-1) type="Field";
	 if (className.indexOf("Class")!=-1) type="Class";
	 String length = (new Integer(this.getByteStopIndex()-getByteStartIndex()+1))
		.toString();
	return type+" "+((DecodeDisplayInterface)this).getName()+" "+length+" bytes";
}
/**
 * Determines which ranges of Strings in the description are actually the names of
 * other java Objects.  These character ranges can be used to mark the original 
 * Description String as hyperlinks or change the color when there are java objects
 * present.
 * 
 * @return Vector - contains java.awt.Point objects which hold the start and end
 *					position of all the names of java objects in the Description 
 *					String.
 */
public Vector getObjectLinks(){
	return new Vector();
}
/**
 * This method creates a 'source like' String that describes this class object.
 * The String will contain the name of the class, imports, and extensions.
 * 
 * @return String - String describing this class using source syntax.
 */
public String getDescription(){
	return null;
}
/**
 * Analyzes the String which contains the raw form of the internal fully qualified
 * class name to determine if the class is a java primative or another java object.
 * 
 * @param String data - raw form of an internal fully qualified class name.
 * 
 * @return boolean - true if this is a primative, false if it is a java object.
 */
public boolean isPrimative(String data){
	int index = 0;
	while (data.charAt(index)=='[')
		index++;
	if (data.charAt(index)=='L')
		return false;
	return true;
}
/**
 * Returns a Vector containing all of the Entry objects associated with this 
 * Decoded Object.  Each entry will contain information about a specific information
 * that has been decoded.
 * 
 * @return Vector - contains all defined Entry Objects.
 */
public Vector getEntries(){
	if (entri.size()==0)
		return null;
	else return entri;
}
/**
 * Creates a new entry object, and adds it to the list of properties for this 
 * DecodeObject.
 * 
 * @param String desc - description of the property that has been decoded.
 * @param int off - offset of the first byte of this property from the start of the
 *					class file.
 * @param int len - the number of bytes that this property occupies.
 * @param String val - an interpretation of these bytes, into human readable form
 * @param byte[] data - the raw bytes that represent the property.
 */
public void addEntry(String desc, int off, int len, String val, byte[] data){
	Entry en = new Entry();
	en.setDescription(desc);
	en.setOffset(off);
	en.setLength(len);
	en.setValue(val);
	en.setBytes(data);
	entri.addElement(en);
}
/**
 * Serves as a placeholder for decode objects that need to process a flags property.
 * Specifically it is used my Method, Field, and Class object.  This method must be
 * overridden by any subclasses that need to process flags.
 * 
 * @see sample.decoder.ClassObject#getAccessFlags
 * @see sample.decoder.FieldObject#getAccessFlags
 * @see sample.decoder.MethodObject#getAccessFlags
 * 
 * @return Vector - contains human readable Strings that represent the access flags
 *					for this decode object.  In this class, it returns null.
 */
protected Vector getAccessFlags(){
	return null;
}
/**
 * Leverages the getAccessFlags method to obtain a list of human readable access
 * flags, and creates a long, concatonated String of the results.
 * 
 * @return String - String containing all access flags for this DecodeObject.
 */
protected String getAccessFlagsString(){
	Vector acc = getAccessFlags();
	String result = "";
	if ((acc==null)||(acc.size()==0))
		return result;
	else{
		result=(String)acc.firstElement();
		for(int i=1;i<acc.size();i++)
			result+=" "+((String)acc.elementAt(i));
		return result;
	}
}
}