/*
 *
 * sample.decoder.ExceptionTableStructure.java
 *
 * (C) Copyright 1995 - 1999 Microsoft Corporation.  All rights reserved.
 *
*/
package sample.decoder;

import java.util.Vector;
import java.util.Hashtable;
/**
 * Each Exception Table structure describes one exception handler in the code array
 * of a Code attribute.  
 * 
 * @see sample.decoder.CodeAttribute
 *
 * @date	April 28,1998
 * @version	1.1
*/
public class ExceptionTableStructure extends DecodeObject implements DecodeDisplayInterface,
		java.io.Serializable{
	/**
	 * This is a valid index into the code array to an opcode instruction that 
	 * represents the activation point for the exception handler.  The start_pc 
	 * value must be less than the end_pc value.
	 */
	protected byte[] start_pc = new byte[2];
	/**
	 * This is a valid index into the code array, or the length of the code array.
	 * Must be greater than the start_pc value. 
	 */
	protected byte[] end_pc = new byte[2];
	/**
	 * This must be less than the code length, and be a valid index into the code
	 * that is an opcode instruction of the code that will handle the exception.
	 */
	protected byte[] handler_pc = new byte[2];
	/**
	 * This must be a non-zero index into the Constant Pool to a Constant Class 
	 * which contains the type of exception that this handler deals with.
	 */
	protected byte[] catch_type = new byte[2];
	/**
	 * The actual Constant Pool entry that is the catch type.
	 */
	protected ConstantClassObject catchType = null;
/**
 * This method is a member of the DecodeDisplayInterface and is used to 
 * return the object name and the attribute name.  Subclasses of this object
 * should over ride this method to provide more detailed information.
 *
 * @see sample.decoder.DecodeDisplayInterface
 *
 * @returns String name representing this object and its attribute type.
*/
public String getName(){
	return "Exception Table";
}
/**
 * Required by the DecodeDisplayInterface, returns any Vectors, or subobjects
 * which will need to be displayed as a subobject in a GUI.  The Hashtable is
 * indexed by a String that is to be used as a display name, the contents
 * will be either a Vector of Strings/Objects or a Decode Object.
 *
 * @returns Hashtable of subobjects/structures to be displayed as seperate
 *				nodes in the GUI.
*/
public Hashtable getChildren(){
	return new Hashtable();
}
/**
 * After the initial decoding phase is complete, where all of the bytes have
 * been moved to their appropriate places, this method retrieves any 
 * referenced Constant Pool objects and stores the information with the 
 * object containing the refernece.
 * 
 * @param Vector data - These are the ConstantObjects that were created from the 
 *						constant pool structures in the class file.
*/
public void link(Vector data){
	if (Decode.getIntegerValue(catch_type)!=0)
	catchType = (ConstantClassObject)data.elementAt(Decode.getIntegerValue(catch_type));
}
/**
 * This is the workhorse for the object, it decodes an instance of itself starting at the
 * specified byte index into the stream provided.  This was created as a static method
 * rather than a constructor, as there may be times when this method returns a more
 * specialized object which is a subclass of itself.
 *
 * @param byte[] stream -  entire bytes of the class that is being decoded
 * @param int index - this is the index into the stream for the starting byte.
 * @param Vector constantPool - these are the constantPool objects, any index into the
 *								constant pool will be resolved by retrieving the object 
 *								at the required instance from this Vector.
 *
 * @returns Object - This is the decoded object which is a populated instance of this 
 *					class.
*/
public static Object getInstance(byte[] stream, int index,Vector constantPool){
        ExceptionTableStructure xconst = new ExceptionTableStructure();
		xconst.setByteStartIndex(index);
		xconst.start_pc[0]=stream[index++];
		xconst.start_pc[1]=stream[index++];
		xconst.addEntry("Start PC",index-2,2,""+Decode.getIntegerValue(xconst.start_pc),
						xconst.start_pc);
		xconst.end_pc[0]=stream[index++];
		xconst.end_pc[1]=stream[index++];
		xconst.addEntry("End PC",index-2,2,""+Decode.getIntegerValue(xconst.end_pc),
						xconst.end_pc);
		xconst.handler_pc[0]=stream[index++];
		xconst.handler_pc[1]=stream[index++];
		xconst.addEntry("Handler PC",index-2,2,""+Decode.getIntegerValue(xconst.handler_pc),
						xconst.handler_pc);
		xconst.catch_type[0]=stream[index++];
		xconst.catch_type[1]=stream[index++];
		xconst.addEntry("Catch Type",index-2,2,""+Decode.getIntegerValue(xconst.catch_type),
						xconst.catch_type);
		xconst.setByteStopIndex(index-1);
		xconst.link(constantPool);
        return xconst;
}
/**
 * Generic method for writting debug information to the screen.  By having all
 * methods in this class send their log information here, there is a central
 * point of change for turning debuggin off, or sending it to a differnet 
 * location.
 *
 * @param String data - message to be displayed.
*/
public synchronized void log(String data){
//	System.out.println(data);
}
}