/*
 *
 * sample.viewer.RichEdit201.java
 *
 * (C) Copyright 1995 - 1999 Microsoft Corporation.  All rights reserved.
 *
*/
package sample.viewer;

import com.ms.wfc.core.*;
import com.ms.wfc.win32.*;
import com.ms.win32.*;
import com.ms.wfc.ui.RichEdit;
import com.ms.wfc.ui.Control;
import com.ms.wfc.ui.CreateParams;
/**
 * This implementation extends the RichEdit1.0 functionality by using the 
 * RichEdit2.0 dll, and exposing java methods for creating Hyperlinks, bolding
 * and underlining as well as others.
 * 
 * @see com.ms.wfc.ui.RichEdit 
 * 
 * @version 1.0a
 */
public class RichEdit201 extends RichEdit {
	/**
	 * This is an ID that will be used to uniquely identify this RichEdit201 Control
	 * amongst others.
	 */
	public final String id = (new Double(Math.random())).toString();
	/**
	 * Object used to house the multicast delegate that will be used for broadcasting
	 * any EN_LINK notifications received to registered listeners.
	 */
	private static final Object EVENT_LINK = new Object();
	/**
	 * Handle to the RichEdit 20 dll that will be loaded for construction of the
	 * RichEdit201 control.
	 */
	static int dllHandle = Windows.LoadLibrary("RichEd20.DLL");
/**
 * Overrides the default richEdit to add support for the RichEdit20A
 * control to be created instead of the default Rich Edit version 1.0
 * control.
 * 
 * @return CreateParams - parameters used during the creation of this 
 *					control.
 */
	protected CreateParams getCreateParams() {
		CreateParams cp = super.getCreateParams();
		cp.className = "RichEdit20A";
        return cp;
	}
/**
 * Registers the specified link event handler with our multicast delegate, ensuring 
 * that the link handler will receive future EN_LINK notifications.
 */
	public void addOnLink(LinkEventHandler value){
		addEventHandler(EVENT_LINK,value);
	}
/**
 * Returns the current event handler for the EN_LINK event.
 *
 * @see #addOnLink
 *  
 * @return LinkEventHandler - Delegator that will handle the EN_LNK event.
 */
	public LinkEventHandler getOnLink(){
		return (LinkEventHandler)getEventHandler(EVENT_LINK);
	}
/**
 * Given a line number, this method will determine the correct character offset 
 * (such as in a RichEdit201.getText() call) for the start of the specified line.
 * 
 * @param int lineIndex - can be any valid line number (0 based)
 * 
 * @return int - the character offset for the start of this line in the RichEdit201.
 */
	public int getLineOffset(int lineIndex){
		return sendMessage(win.EM_LINEINDEX,lineIndex,0);	
	}
/**
 * Fires an event when the user selects a region defined as a link.
 * 
 * @see #setOnLink
 * @see #getOnLink
 * 
 * @param LinkEvent e - the Event to be broadcast to registered listeners.
 */
	public void onLink(LinkEvent e){
		fireLinkEvent(EVENT_LINK,e);
	}
/**
 * Uses the multicast delegate to broadcast the event to the registered users
 * 
 * @param Object key - Multicast delegate to broadcast the event
 * @param LinkEvent e - The event that will be broadcast. 
 */
	private void fireLinkEvent(Object key, LinkEvent e){
		((LinkEventHandler)getEventHandler(key)).invoke(this,e);
	}
/**
 * Removes the specified link event handler from the multicast delegate, so that this
 * handler will no longer receive LinkEvents.
 * 
 * @param LinkEventHandler e - LinkEvent listener to remove 
 */
	public void removeLink(LinkEventHandler e){
		removeEventHandler(EVENT_LINK,e);
	}
/**
 * Processes the Hyperlink events (EN_LINK).
 * 
 * @param com.ms.wfc.app.Message m - Windows message to be processed.
 */
   public void wndProc(com.ms.wfc.app.Message m) {
	    switch (m.msg) {
        case Windows.WM_REFLECT + win.WM_NOTIFY:
        
		if (m.hWnd == getHandle()) {
			ENLINK enlink = (ENLINK)m.getLParam(ENLINK.class);
			switch (enlink.nmhdr.code) {
            case win.EN_LINK:
			    LinkEvent le = new LinkEvent();
				le.start=enlink.range.cpMin;
				le.end=enlink.range.cpMax;
				le.id=enlink.msgID;
				onLink(le);
				break;
           }
		}    
		break;
		}
		super.wndProc(m);
	}
/**
 * Sets the event mask for this RichEdit control so that we will receive 
 * EN_LINK notifications.
 * 
 * @param Event e
 */
	protected void onCreateHandle(Event e) {
		super.onCreateHandle(e);
		sendMessage(win.EM_SETEVENTMASK,0,win.ENM_LINK);
	}
/**
 * Sends the Windows message EM_AUTOURLDETECT which turns on automatic
 * URL detection in the document, and turns off any links in the 
 * document that are not of the known types.
 * 
 * *Note - This must be called from onCreateHandle.
 */
	protected void setURLAutoDetect(){
		int res = sendMessage(win.EM_AUTOURLDETECT,1,0);
	}
/**
 * Sets the currently selected area of text in the RichEdit201 control to be bolded.
 */
	public void setBold(){
		com.ms.win32.CHARFORMATA struct = 
		new com.ms.win32.CHARFORMATA();
		struct.dwMask=win.CFM_BOLD;
		struct.dwEffects=win.CFE_BOLD;
		int flags = win.SCF_SELECTION|win.SCF_WORD;
		int res = sendMessage(win.EM_SETCHARFORMAT, flags, struct);		
	}
/**
 * Sets the currently selected area of text in the RichEdit201 control to be unbolded.
 */
	public void setUnBold(){
		com.ms.win32.CHARFORMATA struct = 
		new com.ms.win32.CHARFORMATA();
		struct.dwMask=win.CFM_BOLD;
		int flags = win.SCF_SELECTION|win.SCF_WORD;
		int res = sendMessage(win.EM_SETCHARFORMAT, flags, struct);		
	}
/**
 * Sets the backgroud color of the RichEdit201 control to the specified value.  The 
 * int value should be of the form:
 * 
 *	int color = 0x00bbggrr
 * 
 * @param int color - integer representing the new color of the background
 */
	public void setBackgroundColor(int color){
		com.ms.win32.CHARFORMATA struct = 
		new com.ms.win32.CHARFORMATA();
		struct.dwMask=win.CFM_LINK;
		struct.dwEffects=win.CFE_LINK;
		int flags = win.SCF_SELECTION|win.SCF_WORD;
		int res = sendMessage(win.EM_SETBKGNDCOLOR,0, color);		
	}
/**
 * Sets the currently selected text to be underlined.
 */
	public void setUnderline(){
		com.ms.win32.CHARFORMATA struct = 
		new com.ms.win32.CHARFORMATA();
		struct.dwMask=win.CFM_UNDERLINE;
		struct.dwEffects=win.CFE_UNDERLINE;
		int flags = win.SCF_SELECTION|win.SCF_WORD;
		int res = sendMessage(win.EM_SETCHARFORMAT, flags, struct);
	}
/**
 * Removes the underline from the currently selected text.
 */
	public void setUnUnderline(){
		com.ms.win32.CHARFORMATA struct = 
		new com.ms.win32.CHARFORMATA();
		struct.dwMask=win.CFM_UNDERLINE;
		int flags = win.SCF_SELECTION|win.SCF_WORD;
		int res = sendMessage(win.EM_SETCHARFORMAT, flags, struct);
	}
/**
 * Sets the characters from the specified range to be selected in the RichEdit201 
 * control.
 * 
 * @param int start - character offset to start the selection.
 * @param int stop - character offset to stop the selection region. 
 */
	public void setSelection(int start,int stop){
		com.ms.win32.CHARRANGE range = new com.ms.win32.CHARRANGE();
		range.cpMax=stop;
		range.cpMin=start;
		sendMessage(win.EM_SETSEL,0,range);
	}
/**
 * Sets the selected text color of the RichEdit201 control to the specified value.  
 * The int value should be of the form:
 * 
 *	int color = 0x00bbggrr
 * 
 * @param int color - integer representing the new color of the selected text
 */
	public void setColor(int color){
		com.ms.win32.CHARFORMATA struct = new com.ms.win32.CHARFORMATA();
		struct.dwMask=win.CFM_COLOR;
		struct.crTextColor=color;
		int flags = win.SCF_SELECTION|win.SCF_WORD;
		int res = sendMessage(win.EM_SETCHARFORMAT, flags, struct);	
	}
/**
 * Sets the Currently selected region as a Hyperlink.  This Control will receive
 * EN_LINK notifications when the user manipulates the selected area.
 */
	public void setHyperLink(){
		com.ms.win32.CHARFORMATA struct = 
		new com.ms.win32.CHARFORMATA();
		struct.dwMask=win.CFM_LINK;
		struct.dwEffects=win.CFE_LINK;
		int flags = win.SCF_SELECTION|win.SCF_WORD;
		int res = sendMessage(win.EM_SETCHARFORMAT, flags, struct);
	}
		/**
		 * Exposes the regular RichEdit properties to the VJ Property editor.
		 */
        public static class ClassInfo extends RichEdit.ClassInfo {
        }
}