package SimpComp ;
import com.ms.wfc.core.* ;


/**
 * MyOwnType
 * 
 * Simple class to demonstrate how to use custom properties in controls
 * and to provide custom value editors for those properties
 *  
 */
public class MyOwnType {
    
    /**
     * All possible values of MyOwnType - use one of these 
     */
    public static final MyOwnType TYPE0 = new MyOwnType((short)0);
    public static final MyOwnType TYPE1 = new MyOwnType((short)1);
    public static final MyOwnType TYPE2 = new MyOwnType((short)2);
    public static final MyOwnType TYPE3 = new MyOwnType((short)3);

    /* Private data
    */
    
    // The value stored by MyOwnType instance
    private short m_value = 0;
    
    /* Methods
    */

    /**
     * Render this instance as a user readable string
     * 
     * @return  <b>String</b> A user readable version of the object
     */
    public String toString() {
        
        String string = "" ;
        
        switch (m_value){
            
            case 0: string = "TYPE0" ; break ; 
            case 1: string = "TYPE1" ; break ; 
            case 2: string = "TYPE2" ; break ; 
            case 3: string = "TYPE3" ; break ; 
            default: string = "UNKNOWN VALUE" ;
        }
        
        return string ;
    }

    /* Private methods
    */
    
    /* Constructors are private to prevent creation of anything 
    other than the default set of values (see static finals above)
    */
    private MyOwnType(short aTypeValue) {
        m_value = aTypeValue ;
    }
    
    private MyOwnType() { /* Do Nothing */ }
    

    /**
     * MyOwnType.Editor
     * 
     * Custom ValueEditor to allow instances of this class to be 
     * edited in the properties window
     * 
     * ValueEditors should implement the IValueEditor interace. 
     * ValueEditor provides a default implementation that can be 
     * derived from in order to reduce the amount of duplicate code
     * to be written.
     * 
     * In fact this class could quite easily be serviced by 
     * deriving from FieldsEditor as follows:
     * 
     *      public static class Editor extends FieldsEditor {
     *
     *           public Editor(Class myType) {        
     *               super(myType, MyOwnType.class, null);
     *           }
     *
     *           public int getStyle() {
     *               int style = super.getStyle();
     *               style |= IValueEditor.STYLE_NOEDITABLETEXT;
     *               return style;
     *           }
     *      }
     * 
     * However for the sake of the demonstration this Editor is 
     * actually derived from ValueEditor
     * 
     */
    public static class Editor extends ValueEditor {
        
        /**
         * Constructor - that either takes a Class object or 
         * a parameterless constructor
         * 
         * @see com.ms.wfc.core.IValueEditor
         */
        public Editor(Class valueType) {
            /* Do nothing */
        }

        /**
         * Answer the style of the ValueEditor - this determines what
         * sort of value editor the component manager expects.
         * 
         * In our simple example:
         * 
         * STYLE_VALUES indicates that the value editor implements 
         * the getValues() method. This style typically causes a 
         * property editor to  display a drop-down list for properties 
         * of the value editor's type. 
         * 
         * STYLE_NOEDITABLETEXT indicates that the value editor does 
         * not support accepting arbitrary text in the 
         * getValueFromText() method. It should accept text that is 
         * returned from getTextFromValue(). This style will typically 
         * cause a property editor to disallow typing in the edit box 
         * for properties of the* value editor's type.
         * 
         * 
         * @return  <b>int</b> The control style flag 
         */
        public int getStyle() {
            return STYLE_VALUES | STYLE_NOEDITABLETEXT  ;
        }

        /**
         * Render the value as a string. null is an acceptable return 
         * value
         * 
         * @param  value   The value to render as a string
         * 
         * @return  <b>String</b> The value as a string
         */
        public String getTextFromValue(Object value) {
            return value == null ? "" : value.toString() ;
        }

        /**
         * Converts the given string to an object value
         * 
         * NOTE: This method will not be called with style 
         * STYLE_NOEDITABLETEXT. Remove this style to exercise this 
         * method
         * 
         * @param  text   The string to convert to a value 
         * 
         * @return  <b>Object</b> The string as a value
         */
        public Object getValueFromText(String text) {
            //NOTE: This should check for invalid values
            return new MyOwnType(new Short(text.trim()).shortValue()) ; 
        } 
        
        /**
         * Returns a list of possible values for the value editor's 
         * type. Used to populate the editor combo-box
         * 
         * @return  <b>Object[]</b> Possible values of MyOwnType
         */
        public Object[] getValues() {
            
            MyOwnType[] values =  { MyOwnType.TYPE0
                                    , MyOwnType.TYPE1
                                    , MyOwnType.TYPE2
                                    , MyOwnType.TYPE3
            } ;
            
            return values ;
        } 

        
        /**
         * Returns the fully qualified name of a constant that 
         * represents the given value, or returns null 
         * if there is no constant for the value. 
         * 
         * For instance: "MyOwnType.TYPE0"
         * 
         * @param  value   The object to return as a fully qualified name 
         * 
         * @return  <b>String</b> The name of the value passed in 
         */
        public String getConstantName(Object value) {
            String valueString = value.toString() ;
            return "MyOwnType." + valueString ;
        }
    }
    
}
