/*
 * sample.crypto.DecryptStream.java
 *
 * (C) Copyright 1995 - 1999 Microsoft Corporation.  All rights reserved.
 *
 */
package sample.crypto;

import sample.win32.*;
/**
 * This stream can be used to decrypt data which has been encrypted with the 
 * EncryptStream object.  This method will take a password to use in generating
 * the public key for decrypting the stream, or will read the public key from the
 * start of the incoming stream.  The encryption is performed by using the J/Direct
 * methods to the Win32 Crypto API through the Advapi32 class.
 * 
 * @see sample.crypto.EncryptStream
 * @see sample.win32.Advapi32
 * 
 * @version 1.0
 */
public class DecryptStream extends java.io.InputStream{
        /**
         * This represents the algorithm used to decode the data.
         */
        private static final int ENCRYPT_ALGORITHM = com.ms.win32.win.CALG_RC4;
        /**
         * Block size for the specified algorithm.
         */
        private static final int ENCRYPT_BLOCK_SIZE=1;
        /**
         * Represents a handle to the CSP.
         */
        private int hProv = 0;
        /**
         * Represents a handle to the key to use for the decryption.
         */
        private int hKey = 0;
        /**
         * Password to use in generating the appropriate decryption key.
         */
        protected String szPassword = null;
        /**
         * Stream to read the data from.
         */
        protected java.io.InputStream hSource = null;
/**
 * 
 * @param java.io.InputStream hSource - Stream containing the data to decrypt.
 * 
 * @param String password - This string will be used to generate a crypto key
 *                      to use when decrypting the data contained in the stream. 
 * 
 * @exception java.io.IOException - this may result from the init method.
 */
        public DecryptStream(java.io.InputStream in,String password)throws java.io.IOException{
            hSource = in;
            szPassword = password;
            init();
        }
/**
 * This constructor creates a stream that will have the public key to be used for
 * encryption stored in its first few bytes.  The Crypto APi will initialize the 
 * key for use when reading bytes from this stream.
 * 
 * @param java.io.InputStream in - Stream that the encrypted data will be read from.
 * 
 * @exception java.io.IOException - this may be thrown when attempting to read the
 *                                  initial bytes from the stream which represent
 *                                  the public key to use for the decryption.
 */
        public DecryptStream(java.io.InputStream in)throws java.io.IOException{
            this(in,null);
        }
/**
 * This method simply closes the input stream and releases the context held for the
 * CSP.
 */
        public void close() throws java.io.IOException{
            hSource.close();    
            Advapi32.releaseContext(hProv,0);
            super.close();  
        }
/**
 * Initializes the CSP and either reads the keyblob from the stream to create a 
 * key to be used for decrypting (if the password is null) or creates a public key
 * for decrypting using the password.  The context to the CSP is also acquired in 
 * this method.
 * 
 * @exception java.io.IOException - this may result from a problem when trying to 
 *                  read the Key Blob when no password has been specified.
 */
    protected void init()throws java.io.IOException{
        hProv = Advapi32.acquireContext(null,null,com.ms.win32.win.PROV_RSA_FULL,0);
        if (szPassword == null){
                byte[] size = new byte[4];
                hSource.read(size);
                int dwKeyBlobLen=getInteger(size);
                byte[] pbKeyBlob = new byte[dwKeyBlobLen];  
                hSource.read(pbKeyBlob);
            hKey = Advapi32.importKey(hProv,pbKeyBlob,dwKeyBlobLen,0,0);
        }else{// szPassword is NOT = null
            int hHash = Advapi32.createHash(hProv,com.ms.win32.win.CALG_MD5,0,0);
            byte[] foo = new byte[szPassword.length()];
            for (int i=0;i<foo.length;i++)
                foo[i]=(byte)szPassword.charAt(i);
            Advapi32.hashData(hHash,foo,szPassword.length(),0);
            hKey = Advapi32.deriveKey(hProv,ENCRYPT_ALGORITHM,hHash,0);
            Advapi32.destroyHash(hHash);
        }//end of else
        int dwBlockLen = 1000-1000%ENCRYPT_BLOCK_SIZE;
        int dwBufferLen =dwBlockLen;
    }
/**
 * Reads the next byte of data from the underlying stream, and decodes it using the
 * key created or read from the initialization method.
 * 
 * @return int - the first byte of this int represents the decoded byte from the
 *              input stream.
 * 
 * @exception java.io.IOException - This may result from a problem reading the input
 *                                  stream.
 */
    public int read()throws java.io.IOException{
        byte[] data = new byte[]{(byte)hSource.read()};
        Advapi32.decrypt(hKey,0,false,0,data,1);
        return data[0];
    }
/**
 * This is a utility method that is used to convert the four bytes read from the
 * stream, and determine what integer value those bytes represent.
 * 
 * @param byte[] data - four bytes that represent a java integer.
 * 
 * @return int - the integer calculated using the input bytes.
 */
    public synchronized int getInteger(byte[] data){
        int index=0;
        int b1 = data[index];
        if (b1<0) b1=b1^((255<<16)+(255<<24)+(255<<8));
        int b2 = data[index+1];
        if (b2<0) b2=b2^((255<<16)+(255<<24)+(255<<8));
        int b3 = data[index+2];
        if (b3<0) b3=b3^((255<<16)+(255<<24)+(255<<8));
        int b4 = data[index+3];     
        if (b4<0) b4=b4^((255<<16)+(255<<24)+(255<<8));
        
        return (b1<<0)|
               (b2<<8)|
               (b3<<16)|
               (b4<<24);
    }
/**
 * This main method serves as a command line interface that will take a source file,
 * destination file and an optional key to use for the decoding.
 * 
 * @param String args[] - the command line args to be used in creating the 
 *              appropriate streams, and for passing a password to the constructor.
 */
    public static void main(String args[]){
        if ((args.length!=3)&&(args.length!=2)){
            System.out.println("USAGE: decryptfile <source file> <destination file> <password>");
            System.exit(1);
        }
        try{
        DecryptStream s = null;
        if (args.length==3)
             s = new DecryptStream(new java.io.FileInputStream(args[0]), args[2]);
        else
            s= new DecryptStream(new java.io.FileInputStream(args[0]));
        
        java.io.FileOutputStream f = new java.io.FileOutputStream(args[1]);
        int count=0;
        while((count = s.read())>=0)
            f.write(count);
        
        s.close();
        f.close();
        }catch(Exception ii){
            System.out.println("Exception decoding file "+ii);
        }
    }
}