/*
 * sample.crypto.EncryptStream.java
 *
 * (C) Copyright 1995 - 1999 Microsoft Corporation.  All rights reserved.
 *
 */
package sample.crypto;

import sample.win32.*;
/**
 * Implementation of an Output Steam that can encrypt data based on a given password
 * or generate a random password for the stream, and add the public key to the start.
 * The resulting encoded stream can be decoded with the sample.crypto.DecryptStream
 * class.
 * 
 * @see sample.crypto.DecryptStream
 * @see sample.win32.Advapi32
 * 
 * @version 1.0
 */
public class EncryptStream extends java.io.OutputStream{
        /**
         * Specifies the encryption algorithm to use for the encoding.
         */
        private static final int ENCRYPT_ALGORITHM = com.ms.win32.win.CALG_RC4;
        /**
         * Indicates the block size for the given encoding algorithm.
         */
        private static final int ENCRYPT_BLOCK_SIZE=1;
        /**
         * This is a handle to the CSP.
         */
        private int hProv = 0;
        /**
         * This is a handle to tke Key that will be used for encoding the data.
         */
        private int hKey = 0;
        /**
         * Stream that will receive the encoded data.
         */
        protected java.io.OutputStream hDestination = null;
        /**
         * Password to generate the encoding stream from.  If this is null, a 
         * rondom key will be generated.
         */
        protected String szPassword = null;
/**
 * Creates a Stream that will use a randomly generated key for encoding.
 * 
 * @param java.io.OutputStream out - Stream to receive the encoded data.
 * 
 * @exception java.io.IOException - this may be thrown during the initialization 
 *                      process where the random key blob is written to the stream.
 */
    public EncryptStream(java.io.OutputStream out)throws java.io.IOException{
        this(out,null);
    }
/**
 * Creates a Stream that will create an encryption key to encode the data with 
 * using the specified password.
 * 
 * @param java.io.OutputStream out - Stream to receive the encoded data.
 * @param String password - String to use for generating the encoding key.
 * 
 * @excpetion java.io.IOException - may be thrown during initialization.
 */
    public EncryptStream(java.io.OutputStream out,String password)throws java.io.IOException{
        hDestination=out;
        szPassword = password;
        init();
    }   
/**
 * This will initialize the Stream by creating an encryption key from the specified
 * password, or by generating a random key and writing the associated key blob to 
 * the destination stream.  The CSP context is obtained here and stored in a global
 * variable.
 * 
 * @exception java.io.IOException - may result from a problem writing the random
 *                          keys Blob to the output stream.
 */
    protected void init()throws java.io.IOException{
        // Get handle to the default provider.
        hProv = Advapi32.acquireContext(null,null,com.ms.win32.win.PROV_RSA_FULL,0);
        if (szPassword == null){
            //Create a random session key.
            hKey = Advapi32.genKey(hProv,ENCRYPT_ALGORITHM,
                                   com.ms.win32.win.CRYPT_EXPORTABLE);
            // Get handle to key exchange public key.
            int hXchgKey = Advapi32.getUserKey(hProv,com.ms.win32.win.AT_KEYEXCHANGE);
            // Determine size of the key blob and allocate memory.
            int dwKeyBlobLen = Advapi32.exportKeyLength(hKey,   hXchgKey,
                                com.ms.win32.win.SIMPLEBLOB,0,null);
            byte[] pbKeyBlob = new byte[dwKeyBlobLen];
            
            // Export session key into a simple key blob.
            dwKeyBlobLen = Advapi32.exportKey(hKey,hXchgKey,
                                              com.ms.win32.win.SIMPLEBLOB,
                                       0,pbKeyBlob,dwKeyBlobLen);
            //Release key exchange handle.
            Advapi32.destroyKey(hXchgKey);
            // Write size of key blob to destination file.
            hDestination.write(this.getBytes(dwKeyBlobLen));
            //Write key blob to destination file.
            hDestination.write(pbKeyBlob,0,dwKeyBlobLen);
            hDestination.flush();
        }else { //End of if szPassword = null
            //Encrypt the file with a session key derived from a password.
            // Create a hash object.
            int hHash = Advapi32.createHash(hProv,com.ms.win32.win.CALG_MD5,0,0);
            // Convert String to byte[]
            byte[] foo = new byte[szPassword.length()];
            for (int i=0;i<foo.length;i++)
                foo[i]=(byte)szPassword.charAt(i);
            // Hash in the password data.
            Advapi32.hashData(hHash,foo,szPassword.length(),0);
            // Derive a session key from the hash object.
            hKey = Advapi32.deriveKey(hProv,ENCRYPT_ALGORITHM,hHash,0);
            // Destroy the hash object.
            Advapi32.destroyHash(hHash);
        }//end of else
        // Determine number of bytes to encrypt at a time.  This must be a 
        // multiple of ENCRYPT_BLOCK_SIZE.
        int dwBlockLen = 1000-1000%ENCRYPT_BLOCK_SIZE;
        int dwBufferLen =0;
        
        // Determine the block size.  If a block cipher is used this must have
        // room for an extra block.
        if(ENCRYPT_BLOCK_SIZE>1)
            dwBufferLen = dwBlockLen+ENCRYPT_BLOCK_SIZE;
        else
            dwBufferLen=dwBlockLen;
    }
/**
 * The given byte is encrypted and writen to the specified output stream.
 * 
 * @param int data - int containing a byte that will be encrypted and written to the
 *                  stream.
 * 
 * @exception java.io.IOException - may result form writing to the destination stream
 */
    public void write(int data)throws java.io.IOException{
        byte[] pdata = new byte[]{(byte)data};
        Advapi32.encrypt(hKey,0,false,0,pdata,1, 1);
                hDestination.write(pdata[0]);
    }
/**
 * Closes the output stream, and releases the context to the CSP.
 * 
 * @exception java.io.IOException - may result from problems when closing the stream
 */
    public void close() throws java.io.IOException{
        hDestination.close();
        Advapi32.releaseContext(hProv,0);
        super.close();
    }
/**
 * This is a utility method that is used to convert the given integer into a 4 byte
 * array that can be written to the stream.
 * 
 * @param int data - integer to be converted into a byte array.
 * 
 * @return byte[] - the converted int.
 */
    public static synchronized byte[] getBytes(int data){
        int index=0;
        byte b1 = (byte)data;
        byte b2 = (byte)(data<<8);
        byte b3 = (byte)(data<<16);
        byte b4 = (byte)(data<<24);     
        return new byte[]{b1,b2,b3,b4};
    }
/**
 * This method facilitates command line usage of this class for encoding a given 
 * file and writing the ecoded contents to a specified file location.  The password
 * given will be used to generate the key, or a random key will be generated and
 * written along with the file.
 * 
 * @param String args - command line parameters.
 */
    public static void main(String args[]){
        if ((args.length!=3)&&(args.length!=2)){
            System.out.println("USAGE: encryptfile <source file> <destination file> <password>");
            System.exit(1);
        }
        try{
            EncryptStream e = null;
            if (args.length==3)
                e=new EncryptStream(new java.io.FileOutputStream(args[1]),args[2]);
            else
                e= new EncryptStream(new java.io.FileOutputStream(args[1]));
            java.io.FileInputStream s = new java.io.FileInputStream(args[0]);
            int count=0;
            while ((count=s.read())>=0)
                e.write(count);
            
            s.close();
            e.close();
        }catch(Exception ii){
            System.out.println("Exception encrypting file "+ii);
        }
    }
}