/*
 * sample.crypto.EnumAlgs.java
 *
 * (C) Copyright 1995 - 1999 Microsoft Corporation.  All rights reserved.
 *
 */
package sample.crypto;

import sample.win32.*;
/**
 * Enumerates the Crypto algorithms that have been installed on this machine.  The
 * name of the algorithm, type, number of bits and the Alg ID will be displayed to
 * standard out in a table format.
 * 
 * @see sample.win32.Advapi32
 * @see sample.crypto.InitUser
 * 
 *  @version 1.0
 */
public class EnumAlgs{
        /**
         * Handle to the CSP
         */
        private int hProv = 0;
        /**
         * The display name of the algorithm will be placed in this byte array 
         * when it is retrieved using the Advapi32.getProvParam method.
         */
        byte[] pbData = new byte[1000];
        /**
         * Flags that will be used in the Advapi32.getProvParam call to retrieve
         * an available algorithm.
         */
        int dwFlags;
/**
 * This method enumerates the algorithms available for use in encrypting data, 
 * and calls the print method to output this to the users screen.
 */
    private void init(){
        hProv=Advapi32.acquireContext(null,null,com.ms.win32.win.PROV_RSA_FULL,0);
        // Enumerate the supported algorithms
        for (int i=0;;i++){
            // Set the CRYPT_FIRST flag the first time through the loop.
            if (i==0)
                dwFlags=com.ms.win32.win.CRYPT_FIRST;
            else
                dwFlags=0;

            //Retrieve information about an algorithm.
            try{
                Advapi32.getProvParam(hProv,com.ms.win32.win.PP_ENUMALGS,
                                pbData,1000,dwFlags);
            }catch(com.ms.dll.Win32Exception wI){
                // Determine when all of the algorithms have been enumerated.
                if (!(wI.getErrorCode()==com.ms.win32.win.ERROR_NO_MORE_ITEMS))
                    System.out.println("Error enumerating algorithms "+
                                            wI.getErrorCode()); 
                done();
            }
            // Extract algorithm information from 'pbData' buffer.
            int aiAlgid = convertByte(pbData,0);
            int dwNameLen = convertByte(pbData,8);
            byte[] strBytes = new byte[pbData.length-12];
            for (int j=0;j<strBytes.length;j++)
                strBytes[j]=pbData[j+11];
            // Print information about algorithm.
            print(new StringBuffer(new String(strBytes).trim()).toString(),
                  getAlgType(aiAlgid),
                  convertByte(pbData,4),
                  aiAlgid);
        }
    }
/**
 * Simple method to convert the integer representing the algid into a human 
 * readable String.
 * 
 * @param int aiAlgid - int for of the algid.
 * 
 * @return String - Human readable form.
 */
    protected String getAlgType(int aiAlgid){
        String pszAlgType = "";
        switch(aiAlgid&(7<<13)){
            case com.ms.win32.win.ALG_CLASS_DATA_ENCRYPT: pszAlgType = "Encrypt ";
                                                          break;
            case com.ms.win32.win.ALG_CLASS_HASH: pszAlgType = "Hash ";
                                                  break;
            case com.ms.win32.win.ALG_CLASS_KEY_EXCHANGE: pszAlgType = "Exchange ";
                                                break;
            case com.ms.win32.win.ALG_CLASS_SIGNATURE: pszAlgType = "Signature ";
                                                break;
            default : pszAlgType = "Unknown ";
        }
        return pszAlgType;
    }
/**
 * This print method is used to 'nicely' format the output into a table.
 * 
 * @param String name - name of the algorithm
 * @param String type - type of algorithm
 * @param int bits - number of bits this alg uses.
 * @param int algid - the algid which will be used to encrypt with this algorithm.
 */
    protected void print(String name, String type, int bits, int algid){
        String tab1 = "\t\t";
        if((name.length()/5)>1)
            tab1="\t";
        String tabAlg = "\t\t";
        if (bits>99)
            tabAlg="\t";
        System.out.println("Name:"+name+""+tab1+"Type:"+type+
                    "\tBits:"+bits+""+tabAlg+"Algid:"+Integer.toHexString(algid));
    }
/**
 * Releases the handle to the CSP.
 * 
 * @exception com.ms.dll.Win32Exception - will be thrown if a problem was encountered
 *          while attempting to release the CSP context.  The exception will contain
 *          the corresponding win32 exception code.
 */
    private void done()throws com.ms.dll.Win32Exception{
        //Release CSP handle.
        if(hProv!=0)
            Advapi32.releaseContext(hProv,0);
        System.exit(0);
    }
/**
 * This utility method is used to convert four bytes from a given byte[] array into 
 * the corresponding java integer.
 * 
 * @param byte[] data - Array containing the bytes to use.
 * @param int index - index of the first byte in the array.
 * 
 * @return int - integer representation of the four bytes.
 */
    private int convertByte(byte[] data,int index){
        int b1 = data[index];
        if (b1<0) b1=b1^((255<<16)+(255<<24)+(255<<8));
        int b2 = data[index+1];
        if (b2<0) b2=b2^((255<<16)+(255<<24)+(255<<8));
        int b3 = data[index+2];
        if (b3<0) b3=b3^((255<<16)+(255<<24)+(255<<8));
        int b4 = data[index+3];     
        if (b4<0) b4=b4^((255<<16)+(255<<24)+(255<<8));
        
        return (b1<<0)|
               (b2<<8)|
               (b3<<16)|
               (b4<<24);
    }
/**
 * Simple main method that calls the init method to enumerate the Cypto algarithms 
 * available on the current system.
 * 
 * @param String[] args - These are the command line arguments, they are currently 
 *          ignored.
 */
    public static void main(String args[]){
        new EnumAlgs().init();
    }
}