/*
 * sample.crypto.InitUser
 *
 * (C) Copyright 1995 - 1999 Microsoft Corporation.  All rights reserved.
 *
 */
package sample.crypto;

import sample.win32.*;
/**
 * Creates the default key container if one does not exist.  This should be run 
 * prior to using any of the Crypto streams, or the EnumAlgs application.
 * 
 * The INITUSER sample sets up the default cryptographic client. This sample (or its 
 * equivalent) must be run prior to running the other cryptography samples, so that 
 * they will work correctly.
 * 
 * The following operations are performed:
 * 
 * 1. A default key container is created.
 * 2. A digital signature key pair is created within the key container.
 * 3. A key exchange key pair is created within the key container.
 * 
 * These operations only need to be done once, unless the operating system
 * is reinstalled. If the default key container and key pairs have already
 * been created, this sample performs no operations.
 * 
 * The InitUser sample uses the following CryptoAPI functions:
 * 
 *      acquireContext
 *      getProvParam
 *      genKey
 *      getUserKey
 *      destroyKey
 *      releaseContext
 * 
 * @see sample.crypto.EnumAlgs
 * @see sample.win32.Advapi32
 * 
 * @version 1.0
 */
public class InitUser{
        int hProv = 0;
        int hKey = 0;
        int dwUserNameLen = 0;
        byte[] szUserName = new byte[100];
/**
 * Most of the work required to create the default key container is performed here.
 * Any errors that may result from creating the default key container will be thrown
 * as exceptions, containing the error code fom GetLastError.
 * 
 * @exception com.ms.dll.Win32Exception - exception resulting from a failed 
 *          crypto API call.
 */
    private void init() throws com.ms.dll.Win32Exception{
        //Attempt to acquire a handle to the default key container.
        try{
            hProv = Advapi32.acquireContext(null,
                                    com.ms.win32.win.MS_DEF_PROV_,
                                    com.ms.win32.win.PROV_RSA_FULL,
                                    0);
        }catch(com.ms.dll.Win32Exception ii){
            //Some sort of error occured.
            hProv = createKeyContainer(szUserName, dwUserNameLen);
        }
        try{
            //Attempt to get handle to signature key
            hKey = Advapi32.getUserKey(hProv,
                            com.ms.win32.win.AT_SIGNATURE);
        }catch(com.ms.dll.Win32Exception exp){
            if(exp.getErrorCode()==com.ms.win32.win.NTE_NO_KEY){
                //Create signature key pair.
                System.out.println("Create signature key pair\n");
                hKey = Advapi32.genKey(hProv,
                                com.ms.win32.win.AT_SIGNATURE,
                                0);
                Advapi32.destroyKey(hKey);
            }else{
                System.out.println("Error "+exp.getErrorCode()+
                                   " during CryptGetUserKey!");
                System.exit(1);
            }
        }
        try{
            // Attempt to get handle to exchange key.
            hKey = Advapi32.getUserKey(hProv,com.ms.win32.win.AT_KEYEXCHANGE);
        }catch(java.lang.RuntimeException oioi){
            //Create key exchange pair.
            System.out.println("Create key exchange pair");
            hKey = Advapi32.genKey(hProv,com.ms.win32.win.AT_KEYEXCHANGE,0);
            Advapi32.destroyKey(hKey);
        }
        Advapi32.releaseContext(hProv,0);
        System.out.println("OK");
    }
 /**
  * If the default key container is not present, this method will create one.
  * 
  * @param byte[] szUserName - this is actually a String that will be used to
  *             create the container.
  * @param int dwUserNameLen - unmber of bytes in the byte array that actually
  *             make up the user name.
  */  
    protected int createKeyContainer(byte[] szUserName,int dwUserNameLen){
        //Create default key container.
        hProv = Advapi32.acquireContext(null,
                                    com.ms.win32.win.MS_DEF_PROV_,
                                    com.ms.win32.win.PROV_RSA_FULL,
                                    com.ms.win32.win.CRYPT_NEWKEYSET);
        //Get name of default key container.
        Advapi32.getProvParam(hProv,
                                          com.ms.win32.win.PP_CONTAINER,
                                          szUserName,
                                          dwUserNameLen,
                                          0);
        System.out.println("Create key container "+szUserName);
        return hProv;
    }
/**
 * Simple static method to allow this class to be run as a Java application. All
 * real work is accomplished in the init method.
 * 
 * @param String[] args - command line parameters, these are currently ignored.
 */
    public static void main(String args[]){
        new InitUser().init();
    }
}