/*
 * sample.crypto.SignStream.java
 *
 * (C) Copyright 1995 - 1999 Microsoft Corporation.  All rights reserved.
 *
 */
package sample.crypto;

import sample.win32.*;
/**
 * Creates a signature for the given stream, which will be calculated using the
 * given password.  This stream can then be verified at the recceiving end as being
 * authentic by reading it through a VerifyStream instance, and providing the same
 * password.  The functionality for this class if implemented using the Windows
 * Crypto APi through J/Direct.
 * 
 * @see sample.crypto.VerifyStream
 * 
 * @version 1.0
 */
public class SignStream extends java.io.OutputStream{
        /**
         * Holds the handle to the CSP.
         */
        private int hProv = 0;
        /**
         * Holds the handle to the hash structure that will be used to generate the
         * signature.
         */
        private int hHash = 0;
        /**
         * This is the output stream that will receive the signature (optional)
         */
        protected java.io.OutputStream out = null;
        /**
         * This output Stream will have the same data writen to it, that this stream
         * received.
         */
        protected java.io.OutputStream dataStream = null;
        /**
         * This is the password that will be used to generate the signature.
         */
        protected String szDescription = null;
        /**
         * The calculated length that the signature will take up.
         */
        protected int dwSignatureLen = 0;
        /**
         * Contains the signature created for the stream of data that was received.
         */
        protected byte[] pbSignature = null;
/**
 * This constructor allows the data to be signed to be written to the given stream, 
 * and the signature for that data to be written to another.  The String will provide
 * the information used to create the signature.
 * 
 * @param java.io.OutputStream o  - Stream that will receive the signature.
 * @param java.io.OutputStream data - Stream that will receive the data.
 * @param String p - characters used in generating the signature.
 */     
    public SignStream(java.io.OutputStream o, java.io.OutputStream data, String p){
        out=o;
        dataStream = data;
        szDescription = p;
        init();
    }
/**
 * This constructor allows the data to be signed with the signature for that data 
 * written to the given stream.  The String will provide the information used to 
 * create the signature.
 * 
 * @param java.io.OutputStream o  - Stream that will receive the signature.
 * @param String p - characters used in generating the signature.
 */     
    public SignStream(java.io.OutputStream o, String password){
        out=o;
        szDescription=password;
        init();
    }
/**
 * Retrieves the generated signature, if all of the data has been written.  If not,
 * a Runtime Exception will be thrown.  Note the close() method must be called to 
 * generate the signature.
 * 
 * @return byte[] - this is the generated signature that will be needed to verify
 *                  the contents.
 * 
 * @exception java.lang.RuntimeException - thrown if the signature has not yet been
 *                                      generated.
 */
    public byte[] getSignature() throws RuntimeException{
        if (pbSignature==null)
            throw new RuntimeException("Stream not closed, no signature created");
        return pbSignature;
    }
/**
 * Calling this method will close the streams in use for the signature and the 
 * data if they were used, and will also cause the signature to be generated for
 * the given datastream.  Any attempts to read the signature before this method
 * is called will result in a RuntimeException.
 * 
 * @exception java.io.IOException - any IO error that may result in the process of
 *          writting the signature stream or in closing the streams.
 */
    public void close()throws java.io.IOException{
        this.signData();
        this.writeSig();
        Advapi32.releaseContext(hProv,0);
        super.close();
    }
/**
 * Initializes this instance by getting a handle to the CSP and by creating the 
 * hash object that will be used to generate the signature.  Any problems with 
 * performing these tasks will result in a RuntimeException.
 * 
 * @exception java.lang.RuntimeException - Will be thrown if there was a problem 
 *      getting a handle to the CSP or the Hash.
 */
    public void init()throws RuntimeException{
        // Get handle to the default provider.
        hProv = Advapi32.acquireContext(null,
                                        null,
                                        com.ms.win32.win.PROV_RSA_FULL,
                                        0);
        // Create hash object.
        hHash = Advapi32.createHash(hProv,
                                     com.ms.win32.win.CALG_MD5,
                                     0,
                                     0);
    }
/**
 *  Hash source file, by taking the data written to this stream and adding it to the 
 * Hash object as well as pushing it into the specified data stream.
 * 
 * @param int data - a byte of data that needs to be signed and written to the 
 *      destination stream.
 * 
 * @exception java.io.IOException - may result from writting to the destination 
 *      stream.  
 * @exception java.lang.RuntimeException - can be thrown if there was a problem 
 *      adding the byte to the hash.
 */
    public void write(int data)throws java.io.IOException,RuntimeException{
            Advapi32.hashData(hHash,new byte[]{(byte)data},1,0);
        if (this.dataStream!=null)
            dataStream.write(data);
    }
/**
 * Sign hash object and determine size of signature.  
 * 
 * @exception RuntimeException - may be thrown if there was a problem creating
 *      the signature.
 */
    protected void signData()throws RuntimeException{
        //Determine size of signature
        dwSignatureLen = Advapi32.getSignHashLength(hHash,
                                  com.ms.win32.win.AT_SIGNATURE);
        // Allocate memory for 'pbSignature'.
        pbSignature = new byte[dwSignatureLen];
        
        // Sign hash object (with signature key).
        Advapi32.signHash(hHash,
                          com.ms.win32.win.AT_SIGNATURE,
                          szDescription,
                          0,
                          pbSignature,
                          dwSignatureLen);
    }
/**
 * Write signature to signature file.
 * 
 * @exception java.io.IOException - will be thrown if there was a problem writting
 *          to the signature file stream, or if the stream could not be closed.
 */
    protected void writeSig()throws java.io.IOException{
        //Open signature file.
        if (out!=null){
            out.write(pbSignature,0,dwSignatureLen);
            out.close();
        }
    }
/**
 * This method facilitates command line usage of this class for signing a given 
 * file and writing the contents to a specified file location.  The password
 * given will be used to generate the signature.
 * 
 * @param String args - command line parameters.
 */
    public static void main(String args[]){
        if (args.length!=3){
            System.out.println("USAGE: signfile <source file> <signature file> <description>");
            System.exit(1);
        }
        try{
            java.io.FileInputStream f = new java.io.FileInputStream(args[0]);
            SignStream s = new SignStream(new java.io.FileOutputStream(args[1]),
                                      args[2]);
            byte[] buf = new byte[1000];
            int count=0;
            while ((count=f.read(buf))>0)
                s.write(buf,0,count);
            s.flush();
            s.close();
        }catch(Exception ii){
            System.out.println("Signing error "+ii);
        }
    }
}