/*
 * sample.crypto.VerifyStream.java
 *
 * (C) Copyright 1995 - 1999 Microsoft Corporation.  All rights reserved.
 *
 */
package sample.crypto;

import sample.win32.*;
/**
 * Verifies a signature for the given stream, which will be calculated using the
 * given password.  The functionality for this class if implemented using the 
 * Windows Crypto APi through J/Direct.
 * 
 * @see sample.crypto.SignStream
 *
 * @version 1.0
 */
public class VerifyStream extends java.io.InputStream{
        /**
         * Holds the handle to the CSP.
         */
        private int hProv = 0;
        /**
         * Holds the handle to the hash structure that will be used to generate the
         * signature.
         */
        private int hHash = 0;
        /**
         * This is the output stream that will receive the signature (optional)
         */
        protected java.io.InputStream in = null;
        /**
         * This output Stream will have the same data writen to it, that this stream
         * received.
         */
        protected java.io.InputStream other = null;
        /**
         * This is the password that will be used to generate the signature.
         */
        protected String szDescription = null;
        /**
         * Contains the result of the verification process.
         */
        protected boolean verify=false;
        /**
         * This boolean is used to determine when the verification process is complete.
         */
        protected boolean done = false;
/** 
 * This Constructor will use the two streams to read the data that is to be verified
 * and to read the signature that will be used for the verification process.  The
 * String is used to generate the correct verification Key.
 * 
 * @param java.io.InputStream o - Stream containing the signature.
 * @param java.io.InputStream other - Stream containing the data to be verified.
 * @param String pass - String which will be used to create the key.
 */
    public VerifyStream(java.io.InputStream o,java.io.InputStream sig, String pass){
        szDescription = pass;
        in = sig;
        other=o;
        init();
    }
/**
 * Determines the result of the verification process.  This will throw an exception
 * if the incoming stream verification is not yet complete.
 * 
 * @return boolean - result of the verification process.
 * 
 * @exception java.lang.RuntimeException - thrown whenever this method is called 
 *          before verification is complete.
 */
    public boolean verify() throws RuntimeException{
        if (!done)
            throw new RuntimeException("Not done verifying stream");
        return verify;  
    }
    
/**
 * Initializes this class by getting a handle to the default CSP provider and by
 * creating a hash object that will be used for the verification process.
 */
    public void init(){
        // Get handle to the default provider.
        hProv = Advapi32.acquireContext(null,null,com.ms.win32.win.PROV_RSA_FULL,0);
        // Create hash object.
        hHash = Advapi32.createHash(hProv,com.ms.win32.win.CALG_MD5,0,0);
        }
/**
 * The Client of this Stream will use this method to read information from the data
 * source.  As the data is read from the datasource, it is hashed using the Crypto
 * API to verify its contents.  Only when the stream is completely read, will the
 * verification result be available.
 * 
 * @exception java.io.IOException - this will be the result of a problem encured while
 *          attempting to read the underlying data source stream.
 */
    public int read()throws java.io.IOException{
        // Hash source file.
        int count = other.read();
        if (count==-1)
            this.verify = checkSignature();
        else 
            Advapi32.hashData(hHash,new byte[]{(byte)count},1,0);
        return count;
    }
/**
 * Once the entire data stream has been written, this method is called to perform
 * the actual verification.  
 * 
 * @exception java.io.IOException - this may result from reading the signature data
 *          from the signature stream.
 */
    protected boolean checkSignature()throws java.io.IOException{
        // Determine size of signature.
        int dwSignatureLen = Advapi32.getSignHashLength(hHash,
                                                    com.ms.win32.win.AT_SIGNATURE);
        // Allocate memory for 'pbSignature'.
        byte[] pbSignature = new byte[dwSignatureLen];
        // Open signature file.
        int counter  = in.read(pbSignature);
        // Get handle to signature key.
        int hSigPublicKey = Advapi32.getUserKey(hProv,com.ms.win32.win.AT_SIGNATURE);
        // Verify signature.
        boolean d = Advapi32.verifySignature(hHash,pbSignature,dwSignatureLen,
                                 hSigPublicKey,szDescription,0);
        done=true;
        return d;
    }
/**
 * Closes the underlying streams that were used to read the data and the signarure, 
 * as well as releasing the context to the CSP.
 * 
 * @exception java.io.IOException - may result from problems closing the underlying
 *          Streams. 
 */
    public void close()throws java.io.IOException{
        other.close();
        in.close();
        Advapi32.releaseContext(hProv,0);
        in.close();
    }
/**
 * This method facilitates command line usage of this class for verifying a given 
 * file. The password given will be used to generate the signature.
 * 
 * @param String args - command line parameters.
 */
    public static void main(String args[]){
        if (args.length!=3){
            System.out.println("USAGE: signfile <source file> <signature file> <description>");
            System.exit(1);
        }
        try{
            VerifyStream v = new VerifyStream(new java.io.FileInputStream(args[0]),
                                     new java.io.FileInputStream(args[1]),
                                     args[2]);
            while (v.read()>=0)
                ;//NOP
            System.out.println("Verify "+v.verify());
        }catch(Exception ii){
            System.out.println("Error verifying file "+ii);
        }
    }
}